# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence"
__author__ = "Patrick Kunzmann", "Thomas Nevolianis"
__all__ = ["GeneralSequence", "NucleotideSequence", "ProteinSequence"]

from .sequence import Sequence
from .alphabet import LetterAlphabet, AlphabetError, AlphabetMapper
import numpy as np
import copy


class GeneralSequence(Sequence):
    """
    This class allows the creation of a sequence with custom
    :class:`Alphabet` without the need to subclass :class:`Sequence`.
        
    Parameters
    ----------
    alphabet : Alphabet
        The alphabet of this sequence.
    sequence : iterable object, optional
        The symbol sequence, the :class:`Sequence` is initialized with.
        For alphabets containing single letter strings, this parameter
        may also be a :class:`str` object.
        By default the sequence is empty.
    """
        
    def __init__(self, alphabet, sequence=()):
        self._alphabet = alphabet
        super().__init__(sequence)

    def __repr__(self):
        """Represent GeneralSequence as a string for debugging."""
        return f"GeneralSequence(Alphabet({self._alphabet}), " \
               f"[{', '.join([repr(symbol) for symbol in self.symbols])}])"

    def __copy_create__(self):
        return GeneralSequence(self._alphabet)
    
    def get_alphabet(self):
        return self._alphabet
    
    def as_type(self, sequence):
        """
        Convert the :class:`GeneralSequence` into a sequence of another
        :class:`Sequence` type.

        This function simply replaces the sequence code of the given
        sequence with the sequence code of this object.

        Parameters
        ----------
        sequence : Sequence
            The `Sequence` whose sequence code is replaced with the one
            of this object.
            The alphabet must equal or extend the alphabet of this
            object.
        
        Returns
        -------
        sequence : Sequence
            The input `sequence` with replaced sequence code.
        
        Raises
        ------
        AlphabetError
            If the the :class:`Alphabet` of the input `sequence` does
            not extend the :class:`Alphabet` of this sequence.
        """
        if not sequence.get_alphabet().extends(self._alphabet):
            raise AlphabetError(
                f"The alphabet of '{type(sequence).__name__}' "
                f"is not compatible with the alphabet of this sequence"
            )
        sequence.code = self.code
        return sequence

class NucleotideSequence(Sequence):
    """
    Representation of a nucleotide sequence (DNA or RNA).
    
    This class may have one of two different alphabets:
    :attr:`unambiguous_alphabet()` contains only the unambiguous DNA
    letters 'A', 'C', 'G' and 'T'.
    :attr:`ambiguous_alphabet()` uses an extended alphabet for ambiguous 
    letters.
    
    Parameters
    ----------
    sequence : iterable object, optional
        The initial DNA sequence. This may either be a list or a string.
        May take upper or lower case letters.
        By default the sequence is empty.
    ambiguous : bool, optional
        If true, the ambiguous alphabet is used. By default the
        object tries to use the unambiguous alphabet. If this fails due
        ambiguous letters in the sequence, the ambiguous alphabet
        is used.
    """
    
    alphabet_unamb = LetterAlphabet(["A","C","G","T"])
    alphabet_amb   = LetterAlphabet(
        ["A","C","G","T","R","Y","W","S",
         "M","K","H","B","V","D","N"]
    )
    
    compl_symbol_dict = {"A" : "T",
                         "C" : "G",
                         "G" : "C",
                         "T" : "A",
                         "M" : "K",
                         "R" : "Y",
                         "W" : "W",
                         "S" : "S",
                         "Y" : "R",
                         "K" : "M",
                         "V" : "B",
                         "H" : "D",
                         "D" : "H",
                         "B" : "V",
                         "N" : "N"}
    # List comprehension does not work in this scope
    _compl_symbols = []
    for _symbol in alphabet_amb.get_symbols():
        _compl_symbols.append(compl_symbol_dict[_symbol])
    _compl_alphabet_unamb = LetterAlphabet(_compl_symbols)
    _compl_mapper = AlphabetMapper(_compl_alphabet_unamb, alphabet_amb)
    
    def __init__(self, sequence=[], ambiguous=None):
        if isinstance(sequence, str):
            sequence = sequence.upper()
        else:
            sequence = [symbol.upper() for symbol in sequence]
        if ambiguous is None:
            try:
                self._alphabet = NucleotideSequence.alphabet_unamb
                seq_code = self._alphabet.encode_multiple(sequence)
            except AlphabetError:
                self._alphabet = NucleotideSequence.alphabet_amb
                seq_code = self._alphabet.encode_multiple(sequence)
        elif not ambiguous:
            self._alphabet = NucleotideSequence.alphabet_unamb
            seq_code = self._alphabet.encode_multiple(sequence)
        else:
            self._alphabet = NucleotideSequence.alphabet_amb
            seq_code = self._alphabet.encode_multiple(sequence)
        super().__init__()
        self.code = seq_code

    def __repr__(self):
        """Represent NucleotideSequence as a string for debugging."""
        if self._alphabet == NucleotideSequence.alphabet_amb:
            ambiguous = True
        else:
            ambiguous = False
        return f'NucleotideSequence("{"".join(self.symbols)}", ambiguous={ambiguous})'

    def __copy_create__(self):
        if self._alphabet == NucleotideSequence.alphabet_amb:
            seq_copy = NucleotideSequence(ambiguous=True)
        else:
            seq_copy = NucleotideSequence(ambiguous=False)
        return seq_copy
    
    def get_alphabet(self):
        return self._alphabet
    
    def complement(self):
        """
        Get the complement nucleotide sequence.
        
        Returns
        -------
        complement : NucleotideSequence
            The complement sequence.
        
        Examples
        --------
        
        >>> dna_seq = NucleotideSequence("ACGCTT")
        >>> print(dna_seq.complement())
        TGCGAA
        >>> print(dna_seq.reverse().complement())
        AAGCGT
        
        """
        # Interpreting the sequence code of this object in the
        # complementary alphabet gives the complementary symbols
        # In order to get the complementary symbols in the original
        # alphabet, the sequence code is mapped from the complementary
        # alphabet into the original alphabet
        compl_code = NucleotideSequence._compl_mapper[self.code]
        return self.copy(compl_code)
    
    def translate(self, complete=False, codon_table=None, met_start=False):
        """
        Translate the nucleotide sequence into a protein sequence.
        
        If `complete` is true, the entire sequence is translated,
        beginning with the first codon and ending with the last codon,
        even if stop codons occur during the translation.
        
        Otherwise this method returns possible ORFs in the
        sequence, even if not stop codon occurs in an ORF.
        
        Parameters
        ----------
        complete : bool, optional
            If true, the complete sequence is translated. In this case
            the sequence length must be a multiple of 3.
            Otherwise all ORFs are translated. (Default: False)
        codon_table : CodonTable, optional
            The codon table to be used. By default the default table
            will be used
            (NCBI "Standard" table with "ATG" as single start codon).
        met_start : bool, optional
            If true, the translation starts always with a 'methionine',
            even if the start codon codes for another amino acid.
            Otherwise the translation starts with the amino acid
            the codon codes for. Only applies, if `complete` is false.
            (Default: False)
        
        Returns
        -------
        protein : ProteinSequence or list of ProteinSequence
            The translated protein sequence. If `complete` is true,
            only a single :class:`ProteinSequence` is returned. Otherwise
            a list of :class:`ProteinSequence` is returned, which contains
            every ORF.
        pos : list of tuple (int, int)
            Is only returned if `complete` is false. The list contains
            a tuple for each ORF.
            The first element of the tuple is the index of the 
            :class:`NucleotideSequence`, where the translation starts.
            The second element is the exclusive stop index, it
            represents the first nucleotide in the
            :class:`NucleotideSequence` after a stop codon.
        
        Examples
        --------
        
        >>> dna_seq = NucleotideSequence("AATGATGCTATAGAT")
        >>> prot_seq = dna_seq.translate(complete=True)
        >>> print(prot_seq)
        NDAID
        >>> prot_seqs, pos = dna_seq.translate(complete=False)
        >>> for seq in prot_seqs:
        ...    print(seq)
        MML*
        ML*
        
        """
        if self._alphabet == NucleotideSequence.alphabet_amb:
            raise AlphabetError("Translation requires unambiguous alphabet")
        # Determine codon_table
        if codon_table is None:
            # Import at this position to avoid circular import
            from .codon import CodonTable
            codon_table = CodonTable.default_table()
        stop_code = ProteinSequence.alphabet.encode("*")
        met_code  = ProteinSequence.alphabet.encode("M")
        
        if complete:
            if len(self) % 3 != 0:
                raise ValueError("Sequence length needs to be a multiple of 3 "
                                 "for complete translation")
            # Reshape code into (n,3), with n being the amount of codons
            codons = self.code.reshape(-1, 3)
            protein_seq = ProteinSequence()
            protein_seq.code = codon_table.map_codon_codes(codons)
            return protein_seq
        
        else:
            protein_seqs = []
            pos = []
            code = self.code
            # Create all three frames
            for shift in range(3):
                # The frame length is always a multiple of 3
                # If there is a trailing partial codon, remove it
                frame_length = ((len(code) - shift) // 3) * 3
                frame = code[shift : shift+frame_length]
                # Reshape frame into (n,3), with n being the amount of codons
                frame_codons = frame.reshape(-1, 3)
                # At first, translate frame completely
                protein_code = codon_table.map_codon_codes(frame_codons)
                # Iterate over all start codons in this frame
                starts = np.where(codon_table.is_start_codon(frame_codons))[0]
                for start_i in starts:
                    # Protein sequence beginning from start codon
                    code_from_start = protein_code[start_i:]
                    # Get all stop codon positions
                    # relative to 'code_from_start'
                    stops = np.where(code_from_start == stop_code)[0]
                    # Find first stop codon after start codon
                    # Include stop -> stops[0] + 1
                    stop_i = stops[0] + 1 if len(stops) > 0 \
                             else len(code_from_start)
                    code_from_start_to_stop = code_from_start[:stop_i]
                    prot_seq = ProteinSequence()
                    if met_start:
                        # Copy as the slice is edited
                        prot_seq.code = code_from_start_to_stop.copy()
                        prot_seq.code[0] = met_code
                    else:
                        prot_seq.code = code_from_start_to_stop
                    protein_seqs.append(prot_seq)
                    # Codon indices are transformed
                    # to nucleotide sequence indices
                    pos.append((shift + start_i*3, shift + (start_i+stop_i)*3))
            # Sort by start position
            order = np.argsort([start for start, stop in pos])
            pos = [pos[i] for i in order]
            protein_seqs = [protein_seqs[i] for i in order]
            return protein_seqs, pos
    
    @staticmethod
    def unambiguous_alphabet():
        """
        Get the unambiguous nucleotide alphabet containing the symbols
        ``A``,  ``C``,  ``G`` and  ``T``.

        Returns
        -------
        alphabet : LetterAlphabet
            The unambiguous nucleotide alphabet.
        """
        return NucleotideSequence.alphabet_unamb
    
    @staticmethod
    def ambiguous_alphabet():
        """
        Get the ambiguous nucleotide alphabet containing the symbols
        ``A``,  ``C``,  ``G`` and  ``T`` and symbols describing
        ambiguous combinations of these.

        Returns
        -------
        alphabet : LetterAlphabet
            The ambiguous nucleotide alphabet.
        """
        return NucleotideSequence.alphabet_amb


class ProteinSequence(Sequence):
    """
    Representation of a protein sequence.
    
    Furthermore this class offers a conversion of amino acids from
    3-letter code into 1-letter code and vice versa.
    
    Parameters
    ----------
    sequence : iterable object, optional
        The initial protein sequence. This may either be a list or a
        string. May take upper or lower case letters. If a list is
        given, the list elements can be 1-letter or 3-letter amino acid
        representations. By default the sequence is empty.
    
    Notes
    -----
    The :class:`Alphabet` of this :class:`Sequence` class does not
    support selenocysteine.
    Please convert selenocysteine (``U``) into cysteine (``C``)
    or use a custom :class:`Sequence` class, if the differentiation is
    necessary.
    """

    _codon_table = None
    
    alphabet = LetterAlphabet(["A","C","D","E","F","G","H","I","K","L",
                               "M","N","P","Q","R","S","T","V","W","Y",
                               "B","Z","X","*"])

    # Masses are taken from
    # https://web.expasy.org/findmod/findmod_masses.html#AA

    _mol_weight_average = np.array([
         71.0788,  # A
        103.1388,  # C
        115.0886,  # D
        129.1155,  # E
        147.1766,  # F
         57.0519,  # G
        137.1411,  # H
        113.1594,  # I
        128.1741,  # K
        113.1594,  # L
        131.1926,  # M
        114.1038,  # N
         97.1167,  # P
        128.1307,  # Q
        156.1875,  # R
         87.0782,  # S
        101.1051,  # T
         99.1326,  # V
        186.2132,  # W
        163.1760,  # Y
          np.nan,  # B
          np.nan,  # Z
          np.nan,  # X
          np.nan,  # *
    ])

    _mol_weight_monoisotopic = np.array([
         71.03711,  # A
        103.00919,  # C
        115.02694,  # D
        129.04259,  # E
        147.06841,  # F
         57.02146,  # G
        137.05891,  # H
        113.08406,  # I
        128.09496,  # K
        113.08406,  # L
        131.04049,  # M
        114.04293,  # N
         97.05276,  # P
        128.05858,  # Q
        156.10111,  # R
         87.03203,  # S
        101.04768,  # T
         99.06841,  # V
        186.07931,  # W
        163.06333,  # Y
        np.nan,  # B
        np.nan,  # Z
        np.nan,  # X
        np.nan,  # *
    ])

    _dict_1to3 = {"A" : "ALA",
                  "C" : "CYS",
                  "D" : "ASP",
                  "E" : "GLU",
                  "F" : "PHE",
                  "G" : "GLY",
                  "H" : "HIS",
                  "I" : "ILE",
                  "K" : "LYS",
                  "L" : "LEU",
                  "M" : "MET",
                  "N" : "ASN",
                  "P" : "PRO",
                  "Q" : "GLN",
                  "R" : "ARG",
                  "S" : "SER",
                  "T" : "THR",
                  "V" : "VAL",
                  "W" : "TRP",
                  "Y" : "TYR",
                  "B" : "ASX",
                  "Z" : "GLX",
                  "X" : "UNK",
                  "*" : " * "}
    
    _dict_3to1 = {}
    for _key, _value in _dict_1to3.items():
        _dict_3to1[_value] = _key
    _dict_3to1["SEC"] = "C"
    _dict_3to1["MSE"] = "M"
    
    def __init__(self, sequence=()):
        dict_3to1 = ProteinSequence._dict_3to1
        alph = ProteinSequence.alphabet
        # Convert 3-letter codes to single letter codes,
        # if list contains 3-letter codes
        sequence = [dict_3to1[symbol.upper()] if len(symbol) == 3
                    else symbol.upper() for symbol in sequence]
        super().__init__(sequence)

    def __repr__(self):
        """Represent ProteinSequence as a string for debugging."""
        return f'ProteinSequence("{"".join(self.symbols)}")'

    def get_alphabet(self):
        return ProteinSequence.alphabet
    
    def remove_stops(self):
        """
        Remove *stop signals* from the sequence.
        
        Returns
        -------
        no_stop : ProteinSequence
            A copy of this sequence without stop signals.
        """
        stop_code = ProteinSequence.alphabet.encode("*")
        no_stop = self.copy()
        seq_code = no_stop.code
        no_stop.code = seq_code[seq_code != stop_code]
        return no_stop
    
    @staticmethod
    def convert_letter_3to1(symbol):
        """
        Convert a 3-letter to a 1-letter amino acid representation.
        
        Parameters
        ----------
        symbol : string
            3-letter amino acid representation.
        
        Returns
        -------
        convert : string
            1-letter amino acid representation.
        """
        return ProteinSequence._dict_3to1[symbol.upper()]
    
    @staticmethod
    def convert_letter_1to3(symbol):
        """
        Convert a 1-letter to a 3-letter amino acid representation.
        
        Parameters
        ----------
        symbol : string
            1-letter amino acid representation.
        
        Returns
        -------
        convert : string
            3-letter amino acid representation.
        """
        return ProteinSequence._dict_1to3[symbol.upper()]

    def get_molecular_weight(self, monoisotopic=False):
        """
        Calculate the molecular weight of this protein.
        
        Average protein molecular weight is calculated by the addition
        of average isotopic masses of the amino acids
        in the protein and the average isotopic mass of one water
        molecule.

        Returns
        -------
        weight : float
            Molecular weight of the protein represented by the sequence.
            Molecular weight values are given in Dalton (Da).
        """
        if monoisotopic:
            weight = np.sum(self._mol_weight_monoisotopic[self.code]) + 18.015
        else:
            weight = np.sum(self._mol_weight_average[self.code]) + 18.015

        if np.isnan(weight):
            raise ValueError(
                "Sequence contains ambiguous amino acids, "
                "cannot calculate weight"
            )
        return weight



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module contains a convenience function for loading structures from
general structure files.
"""

__name__ = "biotite.structure.io"
__author__ = "Patrick Kunzmann"
__all__ = ["load_structure", "save_structure"]

import os.path
import io
from ..atoms import AtomArray, AtomArrayStack


def load_structure(file_path, template=None, **kwargs):
    """
    Load an :class:`AtomArray` or class`AtomArrayStack` from a structure
    file without the need to manually instantiate a :class:`File`
    object.
    
    Internally this function uses a :class:`File` object, based on the
    file extension.
    Trajectory files furthermore require specification of the `template`
    parameter.
    
    Parameters
    ----------
    file_path : str
        The path to structure file.
    template : AtomArray or AtomArrayStack or file-like object or str, optional
        Only required when reading a trajectory file.
    kwargs
        Additional parameters will be passed to either the
        :func:`get_structure()` or :func:`read()` method of the file
        object.
        This does not affect files given via the `template` parameter.
        The only exception is the `atom_i`, which is applied to the template
        as well if number of atoms do not match.
    
    Returns
    -------
    array : AtomArray or AtomArrayStack
        If the file contains multiple models, an AtomArrayStack is
        returned, otherwise an AtomArray is returned.
    
    Raises
    ------
    ValueError
        If the file format (i.e. the file extension) is unknown.
    TypeError
        If a trajectory file is loaded without specifying the
        `template` parameter.
    """
    # Optionally load template from file
    if isinstance(template, (io.IOBase, str)):
        template = load_structure(template)

    # We only need the suffix here
    _, suffix = os.path.splitext(file_path)
    if suffix == ".pdb":
        from .pdb import PDBFile
        file = PDBFile.read(file_path)
        array = file.get_structure(**kwargs)
        if isinstance(array, AtomArrayStack) and array.stack_depth() == 1:
            # Stack containing only one model -> return as atom array
            return array[0]
        else:
            return array
    elif suffix == ".pdbqt":
        from .pdbqt import PDBQTFile
        file = PDBQTFile.read(file_path)
        array = file.get_structure(**kwargs)
        if isinstance(array, AtomArrayStack) and array.stack_depth() == 1:
            # Stack containing only one model -> return as atom array
            return array[0]
        else:
            return array
    elif suffix == ".cif" or suffix == ".pdbx":
        from .pdbx import PDBxFile, get_structure
        file = PDBxFile.read(file_path)
        array = get_structure(file, **kwargs)
        if isinstance(array, AtomArrayStack) and array.stack_depth() == 1:
            # Stack containing only one model -> return as atom array
            return array[0]
        else:
            return array
    elif suffix == ".gro":
        from .gro import GROFile
        file = GROFile.read(file_path)
        array = file.get_structure(**kwargs)
        if isinstance(array, AtomArrayStack) and array.stack_depth() == 1:
            # Stack containing only one model -> return as atom array
            return array[0]
        else:
            return array
    elif suffix == ".mmtf":
        from .mmtf import MMTFFile, get_structure
        file = MMTFFile.read(file_path)
        array = get_structure(file, **kwargs)
        if isinstance(array, AtomArrayStack) and array.stack_depth() == 1:
            # Stack containing only one model -> return as atom array
            return array[0]
        else:
            return array
    elif suffix == ".npz":
        from .npz import NpzFile
        file = NpzFile.read(file_path)
        array = file.get_structure(**kwargs)
        if isinstance(array, AtomArrayStack) and array.stack_depth() == 1:
            # Stack containing only one model -> return as atom array
            return array[0]
        else:
            return array
    elif suffix == ".mol" or suffix == ".sdf":
        from .mol import MOLFile
        file = MOLFile.read(file_path)
        array = file.get_structure(**kwargs)
        # MOL files only contain a single model
        return array
    elif suffix in [".trr", ".xtc", ".tng", ".dcd", ".netcdf"]:
        if template is None:
            raise TypeError("Template must be specified for trajectory files")
        # filter template for atom ids if it is an unfiltered template
        if "atom_i" in kwargs and template.shape[-1] != len(kwargs["atom_i"]):
            template = template[..., kwargs["atom_i"]]
        from .trr import TRRFile
        from .xtc import XTCFile
        from .tng import TNGFile
        from .dcd import DCDFile
        from .netcdf import NetCDFFile
        if suffix == ".trr":
            traj_file_cls = TRRFile
        if suffix == ".xtc":
            traj_file_cls = XTCFile
        if suffix == ".tng":
            traj_file_cls = TNGFile
        if suffix == ".dcd":
            traj_file_cls = DCDFile
        if suffix == ".netcdf":
            traj_file_cls = NetCDFFile
        file = traj_file_cls.read(file_path, **kwargs)
        return file.get_structure(template)
    else:
        raise ValueError(f"Unknown file format '{suffix}'")


def save_structure(file_path, array, **kwargs):
    """
    Save an :class:`AtomArray` or class`AtomArrayStack` to a structure
    file without the need to manually instantiate a :class:`File`
    object.
    
    Internally this function uses a :class:`File` object, based on the
    file extension.
    
    Parameters
    ----------
    file_path : str
        The path to structure file.
    array : AtomArray or AtomArrayStack
        The structure to be saved.
    kwargs
        Additional parameters will be passed to the respective `set_structure`
        method.

    Raises
    ------
    ValueError
        If the file format (i.e. the file extension) is unknown.
    """
    # We only need the suffix here
    _, suffix = os.path.splitext(file_path)
    if suffix == ".pdb":
        from .pdb import PDBFile
        file = PDBFile()
        file.set_structure(array, **kwargs)
        file.write(file_path)
    elif suffix == ".pdbqt":
        from .pdbqt import PDBQTFile
        file = PDBQTFile()
        file.set_structure(array, **kwargs)
        file.write(file_path)
    elif suffix == ".cif" or suffix == ".pdbx":
        from .pdbx import PDBxFile, set_structure
        file = PDBxFile()
        set_structure(file, array, data_block="STRUCTURE", **kwargs)
        file.write(file_path)
    elif suffix == ".gro":
        from .gro import GROFile
        file = GROFile()
        file.set_structure(array, **kwargs)
        file.write(file_path)
    elif suffix == ".mmtf":
        from .mmtf import MMTFFile, set_structure
        file = MMTFFile()
        set_structure(file, array, **kwargs)
        file.write(file_path)
    elif suffix == ".npz":
        from .npz import NpzFile
        file = NpzFile()
        file.set_structure(array, **kwargs)
        file.write(file_path)
    elif suffix == ".mol" or suffix == ".sdf":
        from .mol import MOLFile
        file = MOLFile()
        file.set_structure(array, **kwargs)
        file.write(file_path)
    elif suffix in [".trr", ".xtc", ".tng", ".dcd", ".netcdf"]:
        from .trr import TRRFile
        from .xtc import XTCFile
        from .tng import TNGFile
        from .dcd import DCDFile
        from .netcdf import NetCDFFile
        if suffix == ".trr":
            traj_file_cls = TRRFile
        if suffix == ".xtc":
            traj_file_cls = XTCFile
        if suffix == ".tng":
            traj_file_cls = TNGFile
        if suffix == ".dcd":
            traj_file_cls = DCDFile
        if suffix == ".netcdf":
            traj_file_cls = NetCDFFile
        file = traj_file_cls()
        file.set_structure(array, **kwargs)
        file.write(file_path)
    else:
        raise ValueError(f"Unknown file format '{suffix}'")


# Helper function to estimate elements from atom names
_elements = [elem.upper() for elem in 
["H", "He", "Li", "Be", "B", "C", "N", "O", "F", "Ne", "Na", "Mg",
"Al", "Si", "P", "S", "Cl", "Ar", "K", "Ca", "Sc", "Ti", "V", "Cr", "Mn", "Fe",
"Co", "Ni", "Cu", "Zn", "Ga", "Ge", "As", "Se", "Br", "Kr", "Rb", "Sr", "Y",
"Zr", "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Sb", "Te",
"I", "Xe", "Cs", "Ba", "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb",
"Dy", "Ho", "Er", "Tm", "Yb", "Lu", "Hf", "Ta", "W", "Re", "Os", "Ir", "Pt",
"Au", "Hg", "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr", "Ra", "Ac", "Th", "Pa",
"U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr", "Rf",
"Db", "Sg", "Bh", "Hs", "Mt", "Ds", "Rg", "Cn", "Nh", "Fl", "Mc", "Lv", "Ts",
"Og"]
]
def _guess_element(atom_name):
    # remove digits (1H -> H)
    elem = "".join([i for i in atom_name if not i.isdigit()])
    elem = elem.upper()

    # Some often used elements for biomolecules
    if elem.startswith("C") or elem.startswith("N") or \
        elem.startswith("O") or elem.startswith("S") or \
        elem.startswith("H"):
        return elem[0]

    # Exactly match element abbreviations
    try:
        return _elements[_elements.index(elem[:2])]
    except ValueError:
        try:
            return _elements[_elements.index(elem[0])]
        except ValueError:
            pass

    return ""
 



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This subpackage is used for reading and writing an :class:`AtomArray` or
:class:`AtomArrayStack` using the Gro format used by the gromacs
software package.
"""

__name__ = "biotite.structure.io.gro"
__author__ = "Daniel Bauer"

from .file import *


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import glob
import io
import itertools
from tempfile import TemporaryFile
import biotite.sequence as seq
import biotite.sequence.io.fastq as fastq
import numpy as np
import os
import os.path
from ..util import data_dir
import pytest

@pytest.mark.parametrize("chars_per_line", [None, 80])
def test_access(chars_per_line):
    path = os.path.join(data_dir("sequence"), "random.fastq")
    file = fastq.FastqFile.read(
        path, offset=33, chars_per_line=chars_per_line
    )
    assert len(file) == 20
    assert list(file.keys()) == [f"Read:{i+1:02d}" for i in range(20)]
    del(file["Read:05"])
    assert len(file) == 19
    assert list(file.keys()) == [f"Read:{i+1:02d}" for i in range(20)
                                 if i+1 != 5]
    for seq_str, scores in file.values():
        assert len(seq_str) == len(scores)
        assert (scores >= 0).all()
    seq_str = "ACTCGGT"
    scores = np.array([10,12,20,11,0,80,42])
    file["test"] = seq_str, scores
    seq_str2, scores2 = file["test"]
    assert seq_str == seq_str2
    assert np.array_equal(scores, scores2)

@pytest.mark.parametrize("chars_per_line", [None, 80])
def test_conversion(chars_per_line):
    path = os.path.join(data_dir("sequence"), "random.fastq")
    fasta_file = fastq.FastqFile.read(
        path, offset=33, chars_per_line=chars_per_line
    )
    ref_content = dict(fasta_file.items())

    fasta_file = fastq.FastqFile(offset=33, chars_per_line=chars_per_line)
    for identifier, (sequence, scores) in ref_content.items():
        fasta_file[identifier] = sequence, scores
    temp = TemporaryFile("w+")
    fasta_file.write(temp)

    temp.seek(0)
    fasta_file = fastq.FastqFile.read(
        temp, offset=33, chars_per_line=chars_per_line
    )
    content = dict(fasta_file.items())
    temp.close()
    
    for identifier in ref_content:
        ref_sequence, ref_scores = ref_content[identifier]
        test_sequence, test_scores = content[identifier]
        assert test_sequence == ref_sequence
        assert np.array_equal(test_scores, ref_scores)

def test_rna_conversion():
    sequence = seq.NucleotideSequence("ACGT")
    scores = np.array([0, 0, 0, 0])
    fastq_file = fastq.FastqFile(offset="Sanger")
    fastq.set_sequence(fastq_file, sequence, scores, "seq1", as_rna=False)
    fastq.set_sequence(fastq_file, sequence, scores, "seq2", as_rna=True)
    assert fastq_file["seq1"][0] == "ACGT" 
    assert fastq_file["seq2"][0] == "ACGU"

@pytest.mark.parametrize(
    "file_name",
    glob.glob(os.path.join(data_dir("sequence"), "*.fastq"))
)
def test_read_iter(file_name):
    ref_dict = dict(fastq.FastqFile.read(file_name, offset="Sanger").items())
    
    test_dict = dict(fastq.FastqFile.read_iter(file_name, offset="Sanger"))

    for (test_id, (test_seq, test_sc)), (ref_id, (ref_seq, ref_sc)) \
        in zip(test_dict.items(), ref_dict.items()):
            assert test_id == ref_id
            assert test_seq == ref_seq
            assert (test_sc == ref_sc).all()

@pytest.mark.parametrize(
    "offset, chars_per_line, n_sequences", itertools.product(
        [33, 42, "Solexa"],
        [None, 80],
        [1, 10]
    )
)
def test_write_iter(offset, chars_per_line, n_sequences):
    """
    Test whether :class:`FastqFile.write()` and
    :class:`FastqFile.write_iter()` produce the same output file for
    random sequences and scores.
    """
    LENGTH_RANGE = (50, 150)
    SCORE_RANGE = (10, 60)

    # Generate random sequences and scores
    np.random.seed(0)
    sequences = []
    scores = []
    for i in range(n_sequences):
        seq_length = np.random.randint(*LENGTH_RANGE)
        code = np.random.randint(
            len(seq.NucleotideSequence.alphabet_unamb),
            size=seq_length
        )
        sequence = seq.NucleotideSequence()
        sequence.code = code
        sequences.append(sequence)
        score = np.random.randint(*SCORE_RANGE, size=seq_length)
        scores.append(score)
    
    fastq_file = fastq.FastqFile(offset, chars_per_line)
    for i, (sequence, score) in enumerate(zip(sequences, scores)):
        identifier = f"seq_{i}"
        fastq_file[identifier] = (str(sequence), score)
    ref_file = io.StringIO()
    fastq_file.write(ref_file)
    
    test_file = io.StringIO()
    fastq.FastqFile.write_iter(
        test_file,
        (
            (f"seq_{i}", (str(sequence), score))
            for i, (sequence, score) in enumerate(zip(sequences, scores))
        ),
        offset, chars_per_line
    )

    assert test_file.getvalue() == ref_file.getvalue()


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import pytest
import numpy as np
import biotite.sequence as seq
import biotite.sequence.align as align
from biotite.sequence.align.statistics import EValueEstimator
from .util import sequences


BACKGROUND = np.array(list({
    "A": 35155,
    "C": 8669,
    "D": 24161,
    "E": 28354,
    "F": 17367,
    "G": 33229,
    "H": 9906,
    "I": 23161,
    "K": 25872,
    "L": 40625,
    "M": 10101,
    "N": 20212,
    "P": 23435,
    "Q": 19208,
    "R": 23105,
    "S": 32070,
    "T": 26311,
    "V": 29012,
    "W": 5990,
    "Y": 14488,
    "B": 0,
    "Z": 0,
    "X": 0,
    "*": 0,
}.values())) / 450431


@pytest.mark.parametrize(
    "matrix_name, gap_penalty, ref_lam, ref_k",
    [
        ("BLOSUM62", (-10000, -10000), 0.318, 0.130),
        ("BLOSUM62", (   -12,     -2), 0.300, 0.090),
        ("BLOSUM62", (    -5,     -5), 0.131, 0.009),
        (  "PAM250", (   -16,     -1), 0.172, 0.018),
    ]
)
def test_distribution_param(matrix_name, gap_penalty, ref_lam, ref_k):
    """
    Check if `EValueEstimator` estimates the extreme value distribution
    parameters correctly by comparing them to the parameters described
    in the original publication by Altschul *et al*.
    """
    SAMPLE_LENGTH = 500
    SAMPLE_SIZE = 1000
    
    alphabet = seq.ProteinSequence.alphabet
    matrix = align.SubstitutionMatrix(alphabet, alphabet, matrix_name)

    np.random.seed(0)
    estimator = align.EValueEstimator.from_samples(
        alphabet, matrix, gap_penalty, BACKGROUND,
        SAMPLE_LENGTH, SAMPLE_SIZE
    )

    # Due to relatively low sample size, expect rather large deviation
    assert estimator.lam == pytest.approx(ref_lam, rel=0.1)
    assert estimator.k == pytest.approx(ref_k, rel=0.6)


def test_evalue():
    """
    Check if the estimated E-values for a given score approximately
    match the number of random sequences with equal or better score via
    sampling.
    Low scores that lead to a rather high E-value are required to get
    a reasonable accuracy.
    """
    TEST_SCORES = [30, 40, 50]
    GAP_PENALTY = (-12, -1)
    N_SAMPLES = 10000
    SEQ_LENGTH = 300

    matrix = align.SubstitutionMatrix.std_protein_matrix()
    estimator = align.EValueEstimator.from_samples(
        seq.ProteinSequence.alphabet, matrix, GAP_PENALTY,
        BACKGROUND
    )

    # Generate large number of alignments of random sequences
    np.random.seed(0)
    random_sequence_code = np.random.choice(
        len(seq.ProteinSequence.alphabet),
        size=(N_SAMPLES, 2, SEQ_LENGTH),
        p=BACKGROUND
    )
    sample_scores = np.zeros(N_SAMPLES, dtype=int)
    for i in range(N_SAMPLES):
        seq1 = seq.ProteinSequence()
        seq2 = seq.ProteinSequence()
        seq1.code = random_sequence_code[i,0]
        seq2.code = random_sequence_code[i,1]
        sample_scores[i] = align.align_optimal(
            seq1, seq2, matrix,
            local=True, gap_penalty=GAP_PENALTY, max_number=1
        )[0].score

    e_values = [
        10 ** estimator.log_evalue(score, SEQ_LENGTH, SEQ_LENGTH * N_SAMPLES)
        for score in TEST_SCORES
    ]
    counts = [
        np.count_nonzero(sample_scores >= score) for score in TEST_SCORES
    ]
    assert e_values == pytest.approx(counts, rel=0.5)


def test_score_scaling(sequences):
    """
    Scaling the substitution scores and gap penalties by a constant
    factor should not influence the obtained E-values.
    Test this by aligning real sequences with a standard and scaled
    scoring scheme and comparing the calculated E-values of these
    alignments.
    """
    SCALING_FACTOR = 1000
    GAP_PENALTY = (-12, -1)
    SEQ_LENGTH = 300

    matrix = align.SubstitutionMatrix.std_protein_matrix()

    np.random.seed(0)
    std_estimator = align.EValueEstimator.from_samples(
        seq.ProteinSequence.alphabet, matrix, GAP_PENALTY,
        BACKGROUND
    )
    scores = [
        align.align_optimal(
            sequences[i], sequences[i+1], matrix, GAP_PENALTY, local=True,
            max_number=1
        )[0].score for i in range(9)
    ]
    std_log_evalues = std_estimator.log_evalue(
        scores, SEQ_LENGTH, SEQ_LENGTH
    )

    scaled_matrix = align.SubstitutionMatrix(
        seq.ProteinSequence.alphabet,
        seq.ProteinSequence.alphabet,
        matrix.score_matrix() * SCALING_FACTOR
    )
    scaled_gap_penalty = (
        GAP_PENALTY[0] * SCALING_FACTOR,
        GAP_PENALTY[1] * SCALING_FACTOR
    )
    scaled_estimator = align.EValueEstimator.from_samples(
        seq.ProteinSequence.alphabet, scaled_matrix, scaled_gap_penalty,
        BACKGROUND
    )
    scores = [
        align.align_optimal(
            sequences[i], sequences[i+1], scaled_matrix, scaled_gap_penalty,
            local=True, max_number=1
        )[0].score for i in range(9)
    ]
    scaled_log_evalues = scaled_estimator.log_evalue(
        scores, SEQ_LENGTH, SEQ_LENGTH
    )

    # Due to relatively low sample size, expect rather large deviation
    assert std_log_evalues.tolist() \
        == pytest.approx(scaled_log_evalues.tolist(), rel=0.2)


def test_invalid_scoring_scheme():
    """
    Check if `from_samples()` raises an exception when the expected
    similarity score between to random symbols is positive.
    """
    alph = seq.ProteinSequence.alphabet
    matrix = align.SubstitutionMatrix(
        alph, alph, np.ones((len(alph), len(alph)), dtype=int)
    )
    # Uniform background frequencies
    freq = np.ones(len(alph))
    
    with pytest.raises(ValueError):
        estimator = EValueEstimator.from_samples(alph, matrix, -10, freq)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from tempfile import TemporaryFile
from os.path import join
import biotite.sequence as seq
import biotite.sequence.io.gff as gff
import biotite.sequence.io.genbank as gb
import numpy as np
import pytest
from ..util import data_dir


@pytest.mark.parametrize(
    "path",
    ["bt_lysozyme.gff3", "gg_avidin.gff3", "ec_bl21.gff3", "sc_chrom1.gff3"]
)
def test_conversion_lowlevel(path):
    """
    Test whether the low-level GFF3 interface can properly read
    a GenBank file and write a file, without data changing.
    """
    gff_file = gff.GFFFile.read(join(data_dir("sequence"), path))
    ref_entries = [entry for entry in gff_file]

    gff_file = gff.GFFFile()
    for entry in ref_entries:
        gff_file.append(*entry)
    temp = TemporaryFile("w+")
    gff_file.write(temp)

    temp.seek(0)
    gff_file = gff.GFFFile.read(temp)
    temp.close()
    test_entries = [field for field in gff_file]
    assert test_entries == ref_entries


@pytest.mark.parametrize(
    "path",
    ["bt_lysozyme.gff3", "gg_avidin.gff3", "ec_bl21.gff3", "sc_chrom1.gff3"]
)
def test_conversion_highlevel(path):
    """
    Test whether the high-level GFF3 interface can properly read
    the features from a GFF3 file and write these properties to a file
    without data changing.
    The 'phase' is tested additionally, since it is not part of a
    `Feature` object.
    """
    gff_file = gff.GFFFile.read(join(data_dir("sequence"), path))
    ref_annot = gff.get_annotation(gff_file)
    ref_phases = []
    for _, _, type, _, _, _, _, phase, _ in gff_file:
        if type == "CDS":
            ref_phases.append(phase)

    gff_file = gff.GFFFile()
    gff.set_annotation(gff_file, ref_annot)
    temp = TemporaryFile("w+")
    gff_file.write(temp)

    temp.seek(0)
    gff_file = gff.GFFFile.read(temp)
    temp.close()
    test_annot = gff.get_annotation(gff_file)
    test_phases = []
    for _, _, type, _, _, _, _, phase, _ in gff_file:
        if type == "CDS":
            test_phases.append(phase)
    
    assert ref_annot == test_annot
    assert test_phases == ref_phases


@pytest.mark.parametrize(
    "path", ["bt_lysozyme.gp", "gg_avidin.gb", "ec_bl21.gb", "sc_chrom1.gb"]
)
def test_genbank_consistency(path):
    """
    Test whether the same annotation (if reasonable) can be read from a
    GFF3 file and a GenBank file.
    """
    gb_file = gb.GenBankFile.read(join(data_dir("sequence"), path))
    ref_annot = gb.get_annotation(gb_file)

    gff_file = gff.GFFFile.read(join(data_dir("sequence"), path[:-3] + ".gff3"))
    test_annot = gff.get_annotation(gff_file)
    
    # Remove qualifiers, since they will be different
    # in GFF3 and GenBank
    ref_annot = seq.Annotation(
        [seq.Feature(feature.key, feature.locs) for feature in ref_annot]
    )
    test_annot = seq.Annotation(
        [seq.Feature(feature.key, feature.locs) for feature in test_annot]
    )
    for feature in test_annot:
        # Only CDS, gene, intron and exon should be equal
        # in GenBank and GFF3
        if feature.key in ["CDS", "gene", "intron", "exon"]:
            try:
                assert feature in test_annot
            except AssertionError:
                print(feature.key)
                for loc in feature.locs:
                    print(loc)
                raise


def test_file_access():
    """
    Test getting, setting, deleting and inserting entries in a GFF3
    file.
    """
    file = gff.GFFFile()
    entry_scaffold = ("ab", "cd", 1, 2, None, None, None, {"Id":"foo"})
    entry = ("a",) + entry_scaffold
    file.append(*entry)
    assert file[0] == entry
    file.append(*(("b",) + entry_scaffold))
    file.insert(1, *(("c",) + entry_scaffold))
    file[1] = ("d",) + entry_scaffold
    file.insert(3, *(("e",) + entry_scaffold))
    del file[2]
    assert [seqid for seqid, _, _, _, _, _, _, _, _ in file] \
        == ["a", "d", "e", ]


def test_entry_indexing():
    """
    Test whether a GFF3 file is indexed correctly based on an artificial
    test file with multiple directives, including '##FASTA'.
    """
    with pytest.warns(UserWarning):
        file = gff.GFFFile.read(
            join(data_dir("sequence"), "indexing_test.gff3")
        )
    assert file._directives == [
        ("directive 1", 1),
        ("directive 2", 2),
        ("directive 3", 7),
        ("FASTA", 8),
    ]
    assert file._entries == [3,4,6]



def test_percent_encoding():
    """
    Test whether percent encoding is working correctly based on an
    artificial test file.
    """
    file = gff.GFFFile.read(join(data_dir("sequence"), "percent_test.gff3"))
    seqid, source, type, start, end, score, strand, phase, attrib \
        = file[0]
    assert seqid == "123,456"
    assert source == "ääh"
    assert type == "regi&n"
    assert attrib == {
        "ID"   : "AnID;AnotherID",
        "Name" : "Ångström",
        "c$l$r": "red\tgreen\tblue"
    }

    file2 = gff.GFFFile()
    file.append(seqid, source, type, start, end, score, strand, phase, attrib)
    assert (seqid, source, type, start, end, score, strand, phase, attrib) \
        == file[0]


def test_error():
    """
    Assert that certain exceptions are raised
    """
    file = gff.GFFFile()
    with pytest.raises(ValueError):
        # 'seqid' beginning with '>' is not legal
        file.append(">xyz", "ab", "cd", 1, 2, None, None, None, {"Id":"foo"})
    with pytest.raises(ValueError):
        # String fields must not be empty
        file.append("", "ab", "cd", 1, 2, None, None, None, {"Id":"foo"})
    with pytest.raises(ValueError):
        # String fields must not be empty
        file.append("xyz", "", "cd", 1, 2, None, None, None, {"Id":"foo"})
    with pytest.raises(ValueError):
        # String fields must not be empty
        file.append("xyz", "ab", "", 1, 2, None, None, None, {"Id":"foo"})
    with pytest.raises(ValueError):
        # Attributes not be empty
        file.append("xyz", "ab", "cd", 1, 2, None, None, None, {})


def test_feature_without_id():
    """
    A feature without 'ID' should raise an error if it has multiple
    locations and consequently multiple entries in the GFF3 file.
    """
    annot = seq.Annotation(
        [seq.Feature(
            key  = "CDS",
            locs = [seq.Location(1,2), seq.Location(4,5)],
            qual = {"some" : "qualifiers"}
        )]
    )
    file = gff.GFFFile()
    with pytest.raises(ValueError):
        gff.set_annotation(file, annot)


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.io.gff"
__author__ = "Patrick Kunzmann"
__all__ = ["GFFFile"]

import copy
import string
from urllib.parse import quote, unquote
import warnings
from ....file import TextFile, InvalidFileError
from ...annotation import Location


# All punctuation characters except
# percent, semicolon, equals, ampersand, comma
_NOT_QUOTED = "".join(
    [char for char in string.punctuation if char not in "%;=&,"]
) + " "


class GFFFile(TextFile):
    """
    This class represents a file in *Generic Feature Format 3*
    (`GFF3 <https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md>`_)
    format.

    Similar to GenBank files, GFF3 files contain information about
    features of a reference sequence, but in a more concise and better
    parsable way.
    However, it does not provide additional meta information.

    This class serves as low-level API for accessing GFF3 files.
    It is used as a sequence of entries, where each entry is defined as
    a non-comment and non-directive line.
    Each entry consists of values corresponding to the 9 columns of
    GFF3:

    ==============  ===============================  ==========================================================
    **seqid**       ``str``                          The ID of the reference sequence
    **source**      ``str``                          Source of the data (e.g. ``Genbank``)
    **type**        ``str``                          Type of the feature (e.g. ``CDS``)
    **start**       ``int``                          Start coordinate of feature on the reference sequence
    **end**         ``int``                          End coordinate of feature on the reference sequence
    **score**       ``float`` or ``None``            Optional score (e.g. an E-value)
    **strand**      ``Location.Strand`` or ``None``  Strand of the feature, ``None`` if feature is not stranded
    **phase**       ``int`` or ``None``              Reading frame shift, ``None`` for non-CDS features
    **attributes**  ``dict``                         Additional properties of the feature
    ==============  ===============================  ==========================================================

    Note that the entry index may not be equal to the line index,
    because GFF3 files can contain comment and directive lines.

    Notes
    -----
    Although the GFF3 specification allows mixing in reference sequence
    data in FASTA format via the ``##FASTA`` directive, this class does
    not support extracting the sequence information.
    The content after the ``##FASTA`` directive is simply ignored.
    Please provide the sequence via a separate file or read the FASTA
    data directly via the :attr:`lines` attribute:
    
    >>> import os.path
    >>> from io import StringIO
    >>> gff_file = GFFFile.read(os.path.join(path_to_sequences, "indexing_test.gff3"))
    >>> fasta_start_index = None
    >>> for directive, line_index in gff_file.directives():
    ...     if directive == "FASTA":
    ...         fasta_start_index = line_index + 1
    >>> fasta_data = StringIO("\\n".join(gff_file.lines[fasta_start_index:]))
    >>> fasta_file = FastaFile.read(fasta_data)
    >>> for seq_string in fasta_file.values():
    ...     print(seq_string[:60] + "...")
    TACGTAGCTAGCTGATCGATGTTGTGTGTATCGATCTAGCTAGCTAGCTGACTACACAAT...

    Examples
    --------
    Reading and editing of an existing GFF3 file:

    >>> import os.path
    >>> gff_file = GFFFile.read(os.path.join(path_to_sequences, "gg_avidin.gff3"))
    >>> # Get content of first entry
    >>> seqid, source, type, start, end, score, strand, phase, attrib = gff_file[0]
    >>> print(seqid)
    AJ311647.1
    >>> print(source)
    EMBL
    >>> print(type)
    region
    >>> print(start)
    1
    >>> print(end)
    1224
    >>> print(score)
    None
    >>> print(strand)
    Strand.FORWARD
    >>> print(phase)
    None
    >>> print(attrib)
    {'ID': 'AJ311647.1:1..1224', 'Dbxref': 'taxon:9031', 'Name': 'Z', 'chromosome': 'Z', 'gbkey': 'Src', 'mol_type': 'genomic DNA'}
    >>> # Edit the first entry: Simply add a score
    >>> score = 1.0
    >>> gff_file[0] = seqid, source, type, start, end, score, strand, phase, attrib
    >>> # Delete first entry
    >>> del gff_file[0]

    Writing a new GFF3 file:

    >>> gff_file = GFFFile()
    >>> gff_file.append_directive("Example directive", "param1", "param2")
    >>> gff_file.append(
    ...     "SomeSeqID", "Biotite", "CDS", 1, 99,
    ...     None, Location.Strand.FORWARD, 0,
    ...     {"ID": "FeatureID", "product":"A protein"}
    ... )
    >>> print(gff_file)   #doctest: +NORMALIZE_WHITESPACE
    ##gff-version 3
    ##Example directive param1 param2
    SomeSeqID   Biotite CDS     1       99      .       +       0       ID=FeatureID;product=A protein
    """
    
    def __init__(self):
        super().__init__()
        # Maps entry indices to line indices
        self._entries = None
        # Stores the directives as (directive text, line index)-tuple
        self._directives = None
        # Stores whether the file has FASTA data
        self._has_fasta = None
        self._index_entries()
        self.append_directive("gff-version", "3")
    
    @classmethod
    def read(cls, file):
        """
        Read a GFF3 file.
        
        Parameters
        ----------
        file : file-like object or str
            The file to be read.
            Alternatively a file path can be supplied.
        
        Returns
        -------
        file_object : GFFFile
            The parsed file.
        """
        file = super().read(file)
        file._index_entries()
        return file
    
    def insert(self, index, seqid, source, type, start, end,
               score, strand, phase, attributes):
        """
        Insert an entry at the given index.
        
        Parameters
        ----------
        index : int
            Index where the entry is inserted.
            If the index is equal to the length of the file, the entry
            is appended at the end of the file. 
        seqid : str
            The ID of the reference sequence.
        source : str
            Source of the data (e.g. ``Genbank``).
        type : str
            Type of the feature (e.g. ``CDS``).
        start : int
            Start coordinate of feature on the reference sequence.
        end : int
            End coordinate of feature on the reference sequence.
        score : float or None
            Optional score (e.g. an E-value).
        strand : Location.Strand or None
            Strand of the feature, ``None`` if feature is not stranded.
        phase : int or None
            Reading frame shift, ``None`` for non-CDS features.
        attributes : dict
            Additional properties of the feature.
        """
        if index == len(self):
            self.append(seqid, source, type, start, end,
                        score, strand, phase, attributes)
        else:
            line_index = self._entries[index]
            line = GFFFile._create_line(
                seqid, source, type, start, end,
                score, strand, phase, attributes
            )
            self.lines.insert(line_index, line)
            self._index_entries()
    
    def append(self, seqid, source, type, start, end,
               score, strand, phase, attributes):
        """
        Append an entry to the end of the file.
        
        Parameters
        ----------
        seqid : str
            The ID of the reference sequence.
        source : str
            Source of the data (e.g. ``Genbank``).
        type : str
            Type of the feature (e.g. ``CDS``).
        start : int
            Start coordinate of feature on the reference sequence.
        end : int
            End coordinate of feature on the reference sequence.
        score : float or None
            Optional score (e.g. an E-value).
        strand : Location.Strand or None
            Strand of the feature, ``None`` if feature is not stranded.
        phase : int or None
            Reading frame shift, ``None`` for non-CDS features.
        attributes : dict
            Additional properties of the feature.
        """
        if self._has_fasta:
            raise NotImplementedError(
                "Cannot append feature entries, "
                "as this file contains additional FASTA data"
            )
        line = GFFFile._create_line(
            seqid, source, type, start, end, score, strand, phase, attributes
        )
        self.lines.append(line)
        # Fast update of entry index by adding last line
        self._entries.append(len(self.lines) - 1)
    
    def append_directive(self, directive, *args):
        """
        Append a directive line to the end of the file.
        
        Parameters
        ----------
        directive : str
            Name of the directive.
        *args : str
            Optional parameters for the directive.
            Each argument is simply appended to the directive, separated
            by a single space character.
        
        Raises
        ------
        NotImplementedError
            If the ``##FASTA`` directive is used, which is not
            supported.
        
        Examples
        --------

        >>> gff_file = GFFFile()
        >>> gff_file.append_directive("Example directive", "param1", "param2")
        >>> print(gff_file)
        ##gff-version 3
        ##Example directive param1 param2
        """
        if directive.startswith("FASTA"):
            raise NotImplementedError(
                "Adding FASTA information is not supported"
            )
        directive_line = "##" + directive + " " + " ".join(args)
        self._directives.append((directive_line[2:], len(self.lines)))
        self.lines.append(directive_line)
    
    def directives(self):
        """
        Get the directives in the file.
        
        Returns
        -------
        directives : list of tuple(str, int)
            A list of directives, sorted by their line order.
            The first element of each tuple is the name of the
            directive (without ``##``), the second element is the index
            of the corresponding line.
        """
        # Sort in line order
        return sorted(self._directives, key=lambda directive: directive[1])
        
    def __setitem__(self, index, item):
        seqid, source, type, start, end, score, strand, phase, attrib = item
        line = GFFFile._create_line(
            seqid, source, type, start, end, score, strand, phase, attrib
        )
        line_index = self._entries[index]
        self.lines[line_index] = line

    
    def __getitem__(self, index):
        if (index >= 0 and  index >= len(self)) or \
           (index <  0 and -index >  len(self)):
                raise IndexError(
                    f"Index {index} is out of range for GFFFile with "
                    f"{len(self)} entries"
                )
        
        line_index = self._entries[index]
        # Columns are tab separated
        s = self.lines[line_index].strip().split("\t")
        if len(s) != 9:
            raise InvalidFileError(f"Expected 9 columns, but got {len(s)}")
        seqid, source, type, start, end, score, strand, phase, attrib = s

        seqid = unquote(seqid)
        source = unquote(source)
        type = unquote(type)
        start = int(start)
        end = int(end)
        score = None if score == "." else float(score)
        if strand == "+":
            strand = Location.Strand.FORWARD
        elif strand == "-":
            strand = Location.Strand.REVERSE
        else:
            strand = None
        phase = None if phase == "." else int(phase)
        attrib = GFFFile._parse_attributes(attrib)

        return seqid, source, type, start, end, score, strand, phase, attrib
    
    def __delitem__(self, index):
        line_index = self._entries[index]
        del self.lines[line_index]
        self._index_entries()
    
    def __len__(self):
        return len(self._entries)
    
    def _index_entries(self):
        """
        Parse the file for comment and directive lines.
        Count these lines cumulatively, so that entry indices can be
        mapped onto line indices.
        Additionally track the line index of directive lines.
        """
        self._directives = []
        # Worst case allocation -> all lines contain actual entries
        self._entries = [None] * len(self.lines)
        self._has_fasta = False
        entry_counter = 0
        for line_i, line in enumerate(self.lines):
            if len(line) == 0 or line[0] == " ":
                # Empty line -> do nothing
                pass
            elif line.startswith("#"):
                # Comment or directive
                if line.startswith("##"):
                    # Directive
                    # Omit the leading '##'
                    self._directives.append((line[2:], line_i))
                    if line[2:] == "FASTA":
                        self._has_fasta = True
                        # This parser does not support bundled FASTA
                        # data
                        warnings.warn(
                            "Biotite does not support FASTA data mixed into "
                            "GFF files, the FASTA data will be ignored"
                        )
                        # To ignore the following FASTA data, stop
                        # parsing at this point
                        break
            else:
                # Actual entry
                self._entries[entry_counter] = line_i
                entry_counter += 1
        # Trim to correct size
        self._entries = self._entries[:entry_counter]

    @staticmethod
    def _create_line(seqid, source, type, start, end,
                     score, strand, phase, attributes):
        """
        Create a line for a newly created entry.
        """
        seqid = quote(seqid.strip(), safe=_NOT_QUOTED) \
                if seqid is not None else "."
        source = quote(source.strip(), safe=_NOT_QUOTED) \
                 if source is not None else "."
        type = type.strip()

        # Perform checks
        if len(seqid) == 0:
            raise ValueError("'seqid' must not be empty")
        if len(source) == 0:
            raise ValueError("'source' must not be empty")
        if len(type) == 0:
            raise ValueError("'type' must not be empty")
        if len(attributes) == 0:
            raise ValueError("'attributes' must not be empty")
        if seqid[0] == ">":
            raise ValueError("'seqid' must not start with '>'")
        
        score = str(score) if score is not None else "."
        if strand == Location.Strand.FORWARD:
            strand = "+"
        elif strand == Location.Strand.REVERSE:
            strand = "-"
        else:
            strand = "."
        phase = str(phase) if phase is not None else "."
        attributes = ";".join(
            [quote(key, safe=_NOT_QUOTED) + "=" + quote(val, safe=_NOT_QUOTED)
             for key, val in attributes.items()]
        )

        return "\t".join(
            [seqid, source, type, str(start), str(end),
             str(score), strand, phase, attributes]
        )
    
    @staticmethod
    def _parse_attributes(attributes):
        """
        Parse the *attributes* string into a dictionary.
        """
        attrib_dict = {}
        attrib_entries = attributes.split(";")
        for entry in attrib_entries:
            compounds = entry.split("=")
            if len(compounds) != 2:
                raise InvalidFileError(
                    f"Attribute entry '{entry}' is invalid"
                )
            key, val = compounds
            attrib_dict[unquote(key)] = unquote(val)
        return attrib_dict


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.structure.io.pdbqt"
__author__ = "Patrick Kunzmann, Daniel Bauer"
__all__ = ["PDBQTFile"]

import warnings
import numpy as np
import networkx as nx
from ....file import TextFile, InvalidFileError
from ...error import BadStructureError
from ...atoms import AtomArray, AtomArrayStack
from ...charges import partial_charges
from ...bonds import BondList, BondType, find_connected, find_rotatable_bonds


PARAMETRIZED_ELEMENTS = [
    "H", "C", "N", "O", "P", "S",
    "F", "CL", "BR", "I",
    "MG", "CA", "MN", "FE", "ZN"
]


class PDBQTFile(TextFile):
    """
    This class represents an *AutoDock* PDBQT file.
    
    This class only provides rudimentary support for reading/writing
    the pure atom information.

    EXPERIMENTAL: Future API changes are probable.
    
    Examples
    --------
    
    Write biotin as flexible ligand into a PDBQT file:
    
    >>> import os.path
    >>> ligand = residue("BTN")
    >>> file = PDBQTFile()
    >>> mask = file.set_structure(ligand, rotatable_bonds="all")
    >>> # Print removed nonpolar hydrogen atoms
    >>> print(ligand[~mask])
    HET         0  BTN H101   H         3.745    1.171    0.974
    HET         0  BTN H102   H         4.071    1.343   -0.767
    HET         0  BTN H91    H         2.802   -0.740   -1.211
    HET         0  BTN H92    H         2.476   -0.912    0.530
    HET         0  BTN H81    H         1.289    1.265    0.523
    HET         0  BTN H82    H         1.616    1.437   -1.218
    HET         0  BTN H71    H         0.346   -0.646   -1.662
    HET         0  BTN H72    H         0.020   -0.818    0.079
    HET         0  BTN H2     H        -0.838    1.576   -1.627
    HET         0  BTN H61    H        -3.797    1.837    1.286
    HET         0  BTN H62    H        -3.367    2.738   -0.205
    HET         0  BTN H5     H        -4.307    0.812   -1.205
    HET         0  BTN H4     H        -2.451   -0.038   -2.252
    >>> print(file)
    ROOT
    HETATM    1 C11  BTN     0       5.089  -0.280   0.173  1.00  0.00     0.258 C 
    HETATM    2 O11  BTN     0       4.956  -1.473   0.030  1.00  0.00    -0.264 OA
    ENDROOT
    BRANCH   1   3
    HETATM    3 O12  BTN     0       6.299   0.233   0.444  1.00  0.00    -0.331 OA
    HETATM   17 HO2  BTN     0       7.034  -0.391   0.517  1.00  0.00     0.221 HD
    ENDBRANCH   1   3
    BRANCH   1   4
    HETATM    4 C10  BTN     0       3.896   0.631   0.039  1.00  0.00     0.105 C 
    BRANCH   4   5
    HETATM    5 C9   BTN     0       2.651  -0.200  -0.276  1.00  0.00     0.010 C 
    BRANCH   5   6
    HETATM    6 C8   BTN     0       1.440   0.725  -0.412  1.00  0.00     0.002 C 
    BRANCH   6   7
    HETATM    7 C7   BTN     0       0.196  -0.106  -0.727  1.00  0.00     0.016 C 
    BRANCH   7   8
    HETATM    8 C2   BTN     0      -1.015   0.819  -0.863  1.00  0.00     0.065 C 
    HETATM    9 S1   BTN     0      -1.419   1.604   0.751  1.00  0.00    -0.154 SA
    HETATM   10 C6   BTN     0      -3.205   1.827   0.371  1.00  0.00     0.090 C 
    HETATM   11 C5   BTN     0      -3.530   0.581  -0.476  1.00  0.00     0.091 C 
    HETATM   12 N1   BTN     0      -3.970  -0.507   0.412  1.00  0.00    -0.239 NA
    HETATM   13 C3   BTN     0      -3.141  -1.549   0.271  1.00  0.00     0.272 C 
    HETATM   14 O3   BTN     0      -3.271  -2.589   0.888  1.00  0.00    -0.259 OA
    HETATM   15 N2   BTN     0      -2.154  -1.343  -0.612  1.00  0.00    -0.239 NA
    HETATM   16 C4   BTN     0      -2.289   0.010  -1.175  1.00  0.00     0.093 C 
    HETATM   18 HN1  BTN     0      -4.738  -0.474   1.004  1.00  0.00     0.132 HD
    HETATM   19 HN2  BTN     0      -1.462  -1.982  -0.843  1.00  0.00     0.132 HD
    ENDBRANCH   7   8
    ENDBRANCH   6   7
    ENDBRANCH   5   6
    ENDBRANCH   4   5
    ENDBRANCH   1   4
    TORSDOF 6
    >>> file.write(os.path.join(path_to_directory, "1l2y_mod.pdb"))
    """

    def get_remarks(self, model=None):
        """
        Get the content of ``REMARKS`` lines.
        
        Parameters
        ----------
        model : int, optional
            If this parameter is given, the function will return a
            string from the remarks corresponding to the given
            model number (starting at 1).
            Negative values are used to index models starting from the
            last model insted of the first model.
            If this parameter is omitted, a list of strings
            containing all models will be returned, even if the
            structure contains only one model.

        Returns
        -------
        lines : str or list of str
            The content of ``REMARKS`` lines, without the leading
            ``'REMARKS'``.
        """
        # Line indices where a new model starts
        model_start_i = np.array([i for i in range(len(self.lines))
                                  if self.lines[i].startswith(("MODEL"))],
                                 dtype=int)
        # Line indices with ATOM or HETATM records
        remark_line_i = np.array([i for i in range(len(self.lines)) if
                                  self.lines[i].startswith("REMARK")],
                                 dtype=int)
        # Structures containing only one model may omit MODEL record
        # In these cases model starting index is set to 0
        if len(model_start_i) == 0:
            model_start_i = np.array([0])
        
        if model is None:
            # Add exclusive end of file
            model_start_i = np.concatenate((model_start_i, [len(self.lines)]))
            model_i = 0
            remarks = []
            for i in range(len(model_start_i) - 1):
                start = model_start_i[i]
                stop  = model_start_i[i+1]
                model_remark_line_i = remark_line_i[
                    (remark_line_i >= start) & (remark_line_i < stop)
                ]
                remarks.append(
                    "\n".join([self.lines[i][7:] for i in model_remark_line_i])
                )
            return remarks
        
        else:
            last_model = len(model_start_i)
            if model == 0:
                raise ValueError("The model index must not be 0")
            # Negative models mean index starting from last model
            model = last_model + model + 1 if model < 0 else model

            if model < last_model:
                line_filter = ( ( remark_line_i >= model_start_i[model-1] ) &
                                ( remark_line_i <  model_start_i[model  ] ) )
            elif model == last_model:
                line_filter = (remark_line_i >= model_start_i[model-1])
            else:
                raise ValueError(
                    f"The file has {last_model} models, "
                    f"the given model {model} does not exist"
                )
            remark_line_i = remark_line_i[line_filter]
            
            # Do not include 'REMARK ' itself -> begin from pos 8
            return "\n".join([self.lines[i][7:] for i in remark_line_i])


    def get_structure(self, model=None):
        """
        Get an :class:`AtomArray` or :class:`AtomArrayStack` from the
        PDBQT file.
        
        Parameters
        ----------
        model : int, optional
            If this parameter is given, the function will return an
            :class:`AtomArray` from the atoms corresponding to the given
            model number (starting at 1).
            Negative values are used to index models starting from the
            last model insted of the first model.
            If this parameter is omitted, an :class:`AtomArrayStack`
            containing all models will be returned, even if the
            structure contains only one model.
        
        Returns
        -------
        array : AtomArray or AtomArrayStack
            The return type depends on the `model` parameter.
        """
        # Line indices where a new model starts
        model_start_i = np.array([i for i in range(len(self.lines))
                                  if self.lines[i].startswith(("MODEL"))],
                                 dtype=int)
        # Line indices with ATOM or HETATM records
        atom_line_i = np.array([i for i in range(len(self.lines)) if
                                self.lines[i].startswith(("ATOM", "HETATM"))],
                               dtype=int)
        # Structures containing only one model may omit MODEL record
        # In these cases model starting index is set to 0
        if len(model_start_i) == 0:
            model_start_i = np.array([0])
        
        if model is None:
            depth = len(model_start_i)
            length = self._get_model_length(model_start_i, atom_line_i)
            array = AtomArrayStack(depth, length)
            # Line indices for annotation determination
            # Annotation is determined from model 1,
            # therefore from ATOM records before second MODEL record
            if len(model_start_i) == 1:
                annot_i = atom_line_i
            else:
                annot_i = atom_line_i[atom_line_i < model_start_i[1]]
            # Line indices for coordinate determination
            coord_i = atom_line_i
        
        else:
            last_model = len(model_start_i)
            if model == 0:
                raise ValueError("The model index must not be 0")
            # Negative models mean index starting from last model
            model = last_model + model + 1 if model < 0 else model

            if model < last_model:
                line_filter = ( ( atom_line_i >= model_start_i[model-1] ) &
                                ( atom_line_i <  model_start_i[model  ] ) )
            elif model == last_model:
                line_filter = (atom_line_i >= model_start_i[model-1])
            else:
                raise ValueError(
                    f"The file has {last_model} models, "
                    f"the given model {model} does not exist"
                )
            annot_i = coord_i = atom_line_i[line_filter]
            array = AtomArray(len(coord_i))
        
        # Save atom IDs for later sorting into the original atom order
        atom_id  = np.zeros(array.array_length(), int)

        # Create annotation arrays
        chain_id  = np.zeros(array.array_length(), array.chain_id.dtype)
        res_id    = np.zeros(array.array_length(), array.res_id.dtype)
        ins_code  = np.zeros(array.array_length(), array.ins_code.dtype)
        res_name  = np.zeros(array.array_length(), array.res_name.dtype)
        hetero    = np.zeros(array.array_length(), array.hetero.dtype)
        atom_name = np.zeros(array.array_length(), array.atom_name.dtype)
        element   = np.zeros(array.array_length(), array.element.dtype)

        # Fill annotation array
        # i is index in array, line_i is line index
        for i, line_i in enumerate(annot_i):
            line = self.lines[line_i]
            
            atom_id[i] = int(line[6:11])
            chain_id[i] = line[21].upper().strip()
            res_id[i] = int(line[22:26])
            ins_code[i] = line[26].strip()
            res_name[i] = line[17:20].strip()
            hetero[i] = (False if line[0:4] == "ATOM" else True)
            atom_name[i] = line[12:16].strip()
            element[i] = line[76:78].strip()
        
        # Add annotation arrays to atom array (stack)
        array.chain_id = chain_id
        array.res_id = res_id
        array.ins_code = ins_code
        array.res_name = res_name
        array.hetero = hetero
        array.atom_name = atom_name
        array.element = element
        
        # Fill in coordinates
        if isinstance(array, AtomArray):
            for i, line_i in enumerate(coord_i):
                line = self.lines[line_i]
                array.coord[i,0] = float(line[30:38])
                array.coord[i,1] = float(line[38:46])
                array.coord[i,2] = float(line[46:54])
                
        elif isinstance(array, AtomArrayStack):
            m = 0
            i = 0
            for line_i in atom_line_i:
                if m < len(model_start_i)-1 and line_i > model_start_i[m+1]:
                    m += 1
                    i = 0
                line = self.lines[line_i]
                array.coord[m,i,0] = float(line[30:38])
                array.coord[m,i,1] = float(line[38:46])
                array.coord[m,i,2] = float(line[46:54])
                i += 1
        
        # Sort into the original atom order
        array = array[..., np.argsort(atom_id)]

        return array
    

    def set_structure(self, atoms, charges=None, atom_types=None,
                      rotatable_bonds=None, root=None, include_torsdof=True):
        """
        Write an :class:`AtomArray` into the PDBQT file.
        
        Parameters
        ----------
        atoms : AtomArray, shape=(n,)
            The atoms to be written into this file.
            Must have an associated :class:`BondList`.
        charges : ndarray, shape=(n,), dtype=float, optional
            Partial charges for each atom in `atoms`.
            By default, the charges are calculated using the PEOE method
            (:func:`partial_charges()`).
        atom_types : ndarray, shape=(n,), dtype="U1", optional
            Custom *AutoDock* atom types for each atom in `atoms`.
        rotatable_bonds : None or 'rigid' or 'all' or BondList, optional
            This parameter describes, how rotatable bonds are handled,
            with respect to ``ROOT``, ``BRANCH`` and ``ENDBRANCH``
            lines.

                - ``None`` - The molecule is handled as rigid receptor:
                  No ``ROOT``, ``BRANCH`` and ``ENDBRANCH`` lines will
                  be written.
                - ``'rigid'`` - The molecule is handled as rigid ligand:
                  Only a ``ROOT`` line will be written.
                - ``'all'`` - The molecule is handled as flexible 
                  ligand:
                  A ``ROOT`` line will be written and all rotatable
                  bonds are included using ``BRANCH`` and ``ENDBRANCH``
                  lines.
                - :class:`BondList` - The molecule is handled as
                  flexible ligand:
                  A ``ROOT`` line will be written and all bonds in the
                  given :class:`BondList` are considered flexible via
                  ``BRANCH`` and ``ENDBRANCH`` lines.
            
        root : int, optional
            Specifies the index of the atom following the ``ROOT`` line.
            Setting the root atom is useful for specifying the *anchor*
            in flexible side chains.
            This parameter has no effect, if `rotatable_bonds` is
            ``None``.
            By default, the first atom is also the root atom.
        include_torsdof : bool, optional
            By default, a ``TORSDOF`` (torsional degrees of freedom)
            record is written at the end of the file.
            By setting this parameter to false, the record is omitted.
        
        Returns
        -------
        mask : ndarray, shape=(n,), dtype=bool
            A boolean mask, that is ``False`` for each atom of the input
            ``atoms``, that was removed due to being a nonpolar
            hydrogen.
        """
        if charges is None:
            charges = partial_charges(atoms)
            charges[np.isnan(charges)] = 0
        else:
            if np.isnan(charges).any():
                raise ValueError("Input charges contain NaN values")
        
        # Get AutoDock atom types and remove nonpolar hydrogen atoms
        atoms, charges, types, mask = convert_atoms(atoms, charges)
        # Overwrite calculated atom types with input atom types
        if atom_types is not None:
            types = atom_types[mask]
        
        if rotatable_bonds is None:
            # No rotatable bonds -> the BondList contains no bonds
            rotatable_bonds = BondList(atoms.bonds.get_atom_count())
            use_root = False
        elif rotatable_bonds == "rigid":
            rotatable_bonds = BondList(atoms.bonds.get_atom_count())
            use_root = True
        elif rotatable_bonds == "all":
            rotatable_bonds = find_rotatable_bonds(atoms.bonds)
            use_root = True
        else:
            if rotatable_bonds.ndim != 2 or rotatable_bonds.shape[1] != 2:
                raise ValueError(
                    "An (nx2) array is expected for rotatable bonds"
                )
            rotatable_bonds = BondList(
                len(mask), np.asarray(rotatable_bonds)
            )[mask]
            use_root = True
        
        if root is None:
            root_index = 0
        else:
            # Find new index of root atom, since the index might have
            # been shifted due to removed atoms
            original_indices = np.arange(len(mask))
            new_indices = original_indices[mask]
            try:
                root_index = np.where(new_indices == root)[0][0]
            except IndexError:
                raise ValueError(
                    "The given root atom index points to an nonpolar hydrogen "
                    "atom, that has been removed"
                )
            # Add bonds of the rigid root to rotatable bonds,
            # as they probably have been filtered out,
            # as the root is probably a terminal atom
            for atom, bond_type in zip(*atoms.bonds.get_bonds(root_index)):
                rotatable_bonds.add_bond(root_index, atom, bond_type)
        
        # Break rotatable bonds
        # for simple branch determination in '_write_atoms()'
        atoms.bonds.remove_bonds(rotatable_bonds)

        hetero = ["ATOM" if e == False else "HETATM" for e in atoms.hetero]
        if "atom_id" in atoms.get_annotation_categories():
            atom_id = atoms.atom_id
        else:
            atom_id = np.arange(1, atoms.array_length()+1)
        occupancy = np.ones(atoms.array_length())
        b_factor = np.zeros(atoms.array_length())

        # Convert rotatable bonds into array for easier handling
        # The bond type is irrelevant from this point on
        rotatable_bonds = rotatable_bonds.as_array()[:,:2]

        self.lines = []
        self._write_atoms(
            atoms, charges, types,
            atom_id, hetero, occupancy, b_factor,
            root_index, rotatable_bonds,
            np.zeros(len(rotatable_bonds), dtype=bool), use_root
        )
        if include_torsdof:
            self.lines.append(f"TORSDOF {len(rotatable_bonds)}")

        return mask
    

    def _write_atoms(self, atoms, charges, types,
                     atom_id, hetero, occupancy, b_factor,
                     root_atom, rotatable_bonds, visited_rotatable_bonds,
                     is_root):
        if len(rotatable_bonds) != 0:
            # Get the indices to atoms of this branch, i.e. a group of
            # atoms that are connected by non-rotatable bonds
            # Use 'find_connected()', since rotatable bonds were removed
            # from the BondList before
            this_branch_indices = find_connected(atoms.bonds, root_atom)
            # The root atom of the branch, i.e. the atom connected by
            # the rotatable bond should always be listed first
            # -> Remove root atom and insert it at the beginning
            this_branch_indices = np.insert(
                this_branch_indices[this_branch_indices != root_atom],
                0,
                root_atom
            )
        else:
            # No rotatable bonds
            # -> all atom are in root i.e. this branch
            this_branch_indices = np.arange(atoms.array_length())
        
        if is_root:
            self.lines.append("ROOT")
        for i in this_branch_indices:
            self.lines.append(
                f"{hetero[i]:6}"
                f"{atom_id[i]:>5d} "
                f"{atoms.atom_name[i]:4} "
                f"{atoms.res_name[i]:3} "
                f"{atoms.chain_id[i]:1}"
                f"{atoms.res_id[i]:>4d}"
                f"{atoms.ins_code[i]:1}   "
                f"{atoms.coord[i,0]:>8.3f}"
                f"{atoms.coord[i,1]:>8.3f}"
                f"{atoms.coord[i,2]:>8.3f}"
                f"{occupancy[i]:>6.2f}"
                f"{b_factor[i]:>6.2f}    "
                f"{charges[i]:>6.3f} "
                f"{types[i]:2}"
            )
        if is_root:
            self.lines.append("ENDROOT")

        if len(rotatable_bonds) == 0:
            # No rotatable bonds -> no branching
            return

        for k, (i, j) in enumerate(rotatable_bonds):
            if visited_rotatable_bonds[k]:
                continue

            # Create a new branch for each rotatable bond,
            # that connects to an atom of this branch 
            if i in this_branch_indices:
                this_br_i = i
                new_br_i = j
            elif j in this_branch_indices:
                this_br_i = j
                new_br_i = i
            else:
                # Rotatable bond does not start from this branch
                continue
            
            # Mark rotatable bond as visited as otherwise branches would
            # be created back and forth over the same rotatable bond and
            # this method would never terminate
            visited_rotatable_bonds[k] = True

            self.lines.append(
                f"BRANCH {atom_id[this_br_i]:>3d} {atom_id[new_br_i]:>3d}"
            )
            self._write_atoms(
                atoms, charges, types,
                atom_id, hetero, occupancy, b_factor,
                # The root atom of the branch
                #is the other atom of the rotatable bond
                new_br_i, rotatable_bonds, visited_rotatable_bonds,
                False
            )
            self.lines.append(
                f"ENDBRANCH {atom_id[this_br_i]:>3d} {atom_id[new_br_i]:>3d}"
            )
    

    def _get_model_length(self, model_start_i, atom_line_i):
        """
        Determine length of models and check that all models
        have equal length.
        """
        n_models = len(model_start_i)
        length = None
        for model_i in range(len(model_start_i)):
            model_start = model_start_i[model_i]
            model_stop = model_start_i[model_i+1] if model_i+1 < n_models \
                            else len(self.lines)
            model_length = np.count_nonzero(
                (atom_line_i >= model_start) & (atom_line_i < model_stop)
            )
            if length is None:
                length = model_length
            if model_length != length:
                raise InvalidFileError(
                    f"Model {model_i+1} has {model_length} atoms, "
                    f"but model 1 has {length} atoms, must be equal"
                )
        return length


def convert_atoms(atoms, charges):
    """
    Convert atoms into *AutoDock* compatible atoms.

    Parameters
    ----------
    atoms : AtomArray
        The atoms to be converted.
    charges : ndarray, dtype=float
        Partial charges for the atoms.
    
    Returns
    -------
    converted_atoms : AtomArray
        The input `atoms`, but with deleted nonpolar hydrogen atoms.
    charges : ndarray, dtype=float
        The input `charges`, but with deleted entries for nonpolar
        hydrogen atoms.
    atom_types : ndarray, dtype="U1"
        The *AutoDock* atom types.
    mask : ndarray, shape=(n,), dtype=bool
        A boolean mask, that is ``False`` for each atom of the input
        ``atoms``, that was removed due to being a nonpolar hydrogen.
    """
    charges = charges.copy()
    all_bonds, all_bond_types = atoms.bonds.get_all_bonds()

    atom_types = np.zeros(atoms.array_length(), dtype="U2")
    hydrogen_removal_mask = np.zeros(atoms.array_length(), dtype=bool)
    for i in range(atoms.array_length()):
        element = atoms.element[i]
        bonded_atoms = all_bonds[i][all_bonds[i] != -1]
        if element == "H":
            if len(bonded_atoms) == 0:
                # Free proton
                atom_types[i] = "H"
            elif len(bonded_atoms) == 1:
                j = bonded_atoms[0]
                bonded_element = atoms.element[j]
                if bonded_element == "C":
                    # Remove hydrogen and add its charge
                    # to charge of bonded carbon
                    charges[j] += charges[i]
                    hydrogen_removal_mask[i] = True
                else:
                    atom_types[i] = "HD"
            else:
                raise BadStructureError(
                    "Structure contains hydrogen with multiple bonds"
                )
        elif element == "C":
            if np.isin(
                all_bond_types[i],
                [BondType.AROMATIC_SINGLE, BondType.AROMATIC_DOUBLE]
            ).any():
                # Aromatic carbon
                atom_types[i] = "A"
            else:
                # Alphatic carbon
                atom_types[i] = "C"
        elif element == "N":
            atom_types[i] = "NA"
        elif element == "O":
            atom_types[i] = "OA"
        elif element == "S":
            atom_types[i] = "SA"
        elif element in PARAMETRIZED_ELEMENTS:
            atom_types[i] = element
        else:
            warnings.warn(
                f"Element {element} is not paramtrized, "
                f"using parameters for hydrogen instead"
            ) 
            atom_types[i] = "H"
    
    mask = ~hydrogen_removal_mask
    return atoms[mask], charges[mask], atom_types[mask], mask


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.sequence.graphics"
__author__ = "Patrick Kunzmann"
__all__ = ["SymbolPlotter", "LetterPlotter", "LetterSimilarityPlotter",
           "LetterTypePlotter",
           "plot_alignment", "plot_alignment_similarity_based",
           "plot_alignment_type_based"]

import abc
import numpy as np
from ...visualize import colors
from .colorschemes import get_color_scheme


class SymbolPlotter(metaclass=abc.ABCMeta):
    """
    Subclasses of this abstract base class define how symbols in
    an alignment are drawn onto an :class:`Axes` object.

    Subclasses must override the :func:`plot_symbol()` method.

    Parameters
    ----------
    axes : Axes
        A *Matplotlib* axes, that is used as plotting area.
    """

    def __init__(self, axes):
        self._axes = axes

    @property
    def axes(self):
        return self._axes

    @abc.abstractmethod
    def plot_symbol(self, bbox, alignment, column_i, seq_i):
        """
        Get the color of a symbol at a specified position in the
        alignment.

        The symbol is specified as position in the alignment's trace
        (``trace[pos_i, seq_i]``).

        Parameters
        ----------
        bbox : Bbox
            The axes area to plot the symbol in
        alignment : Alignment
            The respective alignment.
        column_i : int
            The position index in the trace.
        seq_i : int
            The sequence index in the trace.
        """
        pass


class LetterPlotter(SymbolPlotter, metaclass=abc.ABCMeta):
    """
    This abstract :class:`SymbolPlotter` is the most widely used one.
    Symbols are visualized as character on a colored background box or
    as colored character, if `color_symbols` is set to true.

    Subclasses must override the :class:`get_color()` method.

    Parameters
    ----------
    axes : Axes
        A *Matplotlib* axes, that is used as plotting area.
    color_symbols : bool, optional
        If true, the symbols themselves are colored.
        If false, the symbols are black, and the boxes behind the
        symbols are colored.
    font_size : float, optional
        Font size of the sequence symbols.
    font_param : dict, optional
        Additional parameters that is given to the
        :class:`matplotlib.Text` instance of each symbol.
    """

    def __init__(self, axes, color_symbols=False,
                 font_size=None, font_param=None):
        super().__init__(axes)
        self._color_symbols = color_symbols
        self._font_size = font_size
        self._font_param = font_param if font_param is not None else {}

    def plot_symbol(self, bbox, alignment, column_i, seq_i):
        from matplotlib.patches import Rectangle

        trace = alignment.trace
        if trace[column_i, seq_i] != -1:
            symbol = alignment.sequences[seq_i][trace[column_i, seq_i]]
        else:
            symbol = "-"
        color = self.get_color(alignment, column_i, seq_i)

        box = Rectangle(bbox.p0, bbox.width, bbox.height)
        self.axes.add_patch(box)
        text = self.axes.text(
            bbox.x0 + bbox.width/2, bbox.y0 + bbox.height/2,
            symbol, color="black", ha="center", va="center",
            size=self._font_size, **self._font_param)
        text.set_clip_on(True)

        if self._color_symbols:
            box.set_color("None")
            text.set_color(color)
        else:
            box.set_color(color)

    @abc.abstractmethod
    def get_color(self, alignment, column_i, seq_i):
        """
        Get the color of a symbol at a specified position in the
        alignment.

        The symbol is specified as position in the alignment's trace
        (``trace[pos_i, seq_i]``).

        PROTECTED: Override when inheriting.

        Parameters
        ----------
        alignment : Alignment
            The respective alignment.
        column_i : int
            The position index in the trace.
        seq_i : int
            The sequence index in the trace.

        Returns
        -------
        color : object
            A *Matplotlib* compatible color used for the background
            or the symbol itself at the specifed position
        """
        pass


class LetterSimilarityPlotter(LetterPlotter):
    r"""
    This :class:`SymbolPlotter` colors the symbols based on the
    similarity with the other symbols in the same column.

    The color intensity (or colormap value, respectively) of a symbol
    scales with similarity of the respective symbol to the other symbols
    in the same alignment column.

    Parameters
    ----------
    axes : Axes
        A *Matplotlib* axes, that is used as plotting area.
    matrix : SubstitutionMatrix, optional
        The substitution matrix to use the similarity scores from.
        By default the normalized similarity is 1 for identity and 0
        for non-identity.
    color_symbols : bool, optional
        If true, the symbols themselves are colored.
        If false, the symbols are black, and the boxes behind the
        symbols are colored.
    font_size : float, optional
        Font size of the sequence symbols.
    font_param : dict, optional
        Additional parameters that is given to the
        :class:`matplotlib.Text` instance of each symbol.

    Notes
    -----
    For determination of the color, this a measure called
    *average normalized similarity* is used.

    The *normalized similarity* of one symbol *a* to another symbol *b*
    (both in aphabet *X*) is defined as

    .. math:: S_{norm}(a,b) = \frac{S(a,b) - \min\limits_x(S(a,x))} {\max\limits_x(S(a,x)) - \min\limits_x(S(a,x))}

    .. math:: a,b,x \in X

    where *S(x,y)* is the similarity score of the two symbols
    *x* and *y* described in the substitution matrix.
    The similarity *S(x,-)* is always 0.
    As the normalization is conducted only with respect to *a*,
    the *normalized similarity* is not commutative.

    The *average normalized similarity* of a symbol *a* is
    determined by averaging the normalized similarity over each
    symbol *b*\ :sub:`i` in the same alignment column.

    .. math:: S_{norm,av}(a) = \frac{1}{n-1} \left[\left(\sum\limits_{i=1}^n S_{norm}(a,b_i)\right) - S_{norm}(a,a)\right]

    The normalized similarity of *a* to itself is subtracted,
    because *a* does also occur in *b*\ :sub:`i`.
    """

    def __init__(self, axes, matrix=None, color_symbols=False,
                 font_size=None, font_param=None):

        super().__init__(axes, color_symbols, font_size, font_param)
        if matrix is not None:
            self._matrix = matrix.score_matrix()
        else:
            self._matrix = None
        # Default colormap
        self._cmap = self._generate_colormap(colors["dimgreen"],
                                             self._color_symbols)

    def set_color(self, color=None, cmap=None):
        """
        Set the alignemnt colors used for plotting.

        This function takes either a color or a colormap.

        Parameters
        ----------
        color : tuple or str, optional
            A *Matplotlib* compatible color.
            If this parameter is given, the box color in an interpolated
            value between white and the given color,
            or, if `color_symbols` is set, between the given color and
            black.
            The interpolation percentage is given by the normalized
            similarity.
        cmap : Colormap, optional
            The boxes (or symbols, if `color_symbols` is set) are
            colored based on the normalized similarity value on the
            given Colormap.
        """
        from matplotlib import cm

        if color is None and cmap is None:
            raise ValueError("Either color or colormap must be set")
        elif color is not None:
            self._cmap = self._generate_colormap(color, self._color_symbols)
        else:
            # cmap is not None
            if isinstance(cmap, str):
                self._cmap = cm.get_cmap(cmap)
            else:
                # cmap is a colormap
                self._cmap = cmap

    def get_color(self, alignment, column_i, seq_i):
        # Calculate average normalize similarity
        index1 = alignment.trace[column_i, seq_i]
        if index1 == -1:
            similarity = 0
        else:
            code1 = alignment.sequences[seq_i].code[index1]
            similarities = np.zeros(alignment.trace.shape[1])
            for i in range(alignment.trace.shape[1]):
                index2 = alignment.trace[column_i, i]
                if index2 == -1:
                    similarities[i] = 0
                else:
                    code2 = alignment.sequences[i].code[index2]
                    similarities[i] = self._get_similarity(self._matrix,
                                                           code1, code2)
            # Delete self-similarity
            similarities = np.delete(similarities, seq_i)
            similarity = np.average(similarities)
        return self._cmap(similarity)

    def _get_similarity(self, matrix, code1, code2):
        if matrix is None:
            return 1 if code1 == code2 else 0
        else:
            sim = matrix[code1, code2]
            # Normalize (range 0.0 - 1.0)
            min_sim = np.min(matrix[code1])
            max_sim = np.max(matrix[code1])
            sim = (sim - min_sim) / (max_sim - min_sim)
            return sim

    @staticmethod
    def _generate_colormap(color, to_black):
        from matplotlib.colors import ListedColormap, to_rgb

        color = to_rgb(color)
        if to_black:
            # From color to black
            cmap_val = np.stack(
                [np.interp(np.linspace(0, 1, 100), [0, 1], [color[i], 0])
                 for i in range(len(color))]
            ).transpose()
        else:
            # From white to color
            cmap_val = np.stack(
                [np.interp(np.linspace(0, 1, 100), [0, 1], [1, color[i]])
                 for i in range(len(color))]
            ).transpose()
        return ListedColormap(cmap_val)


class LetterTypePlotter(LetterPlotter):
    """
    This `SymbolPloter` colors each symbol based on the general
    color of that symbol defined by a color scheme.

    EXPERIMENTAL: Future API changes are probable.

    Parameters
    ----------
    axes : Axes
        A *Matplotlib* axes, that is used as plotting area.
    alphabet : Alphabet
        The alphabet of the alignment(s) to be plotted
    color_scheme : str or list of (tuple or str), optional
        Either a valid color scheme name
        (e.g. ``"rainbow"``, ``"clustalx"``, ``blossom``, etc.)
        or a list of *Matplotlib* compatible colors.
        The list length must be at least as long as the
        length of the alphabet used by the sequences.
    color_symbols : bool, optional
        If true, the symbols themselves are colored.
        If false, the symbols are black, and the boxes behind the
        symbols are colored.
    font_size : float, optional
        Font size of the sequence symbols.
    font_param : dict, optional
        Additional parameters that is given to the
        :class:`matplotlib.Text` instance of each symbol.
    """

    def __init__(self, axes, alphabet, color_scheme=None, color_symbols=False,
                 font_size=None, font_param=None):
        super().__init__(axes, color_symbols, font_size, font_param)

        if color_scheme is None:
            self._colors = get_color_scheme("rainbow", alphabet)
        elif isinstance(color_scheme, str):
            self._colors = get_color_scheme(color_scheme, alphabet)
        else:
            self._colors = color_scheme

    def get_color(self, alignment, column_i, seq_i):
        index = alignment.trace[column_i, seq_i]
        if index == -1:
            # Gaps are white
            return (1, 1, 1)
        code = alignment.sequences[seq_i].code[index]
        return self._colors[code]


def plot_alignment(axes, alignment, symbol_plotter, symbols_per_line=50,
                   show_numbers=False, number_size=None, number_functions=None,
                   labels=None, label_size=None,
                   show_line_position=False,
                   spacing=1, symbol_spacing=None):
    """
    Plot a pairwise or multiple sequence alignment.

    The output is similar to a string representation of an
    :class:`Alignment`, but with enhanced styling, symbol coloring and
    optional sequence labels and sequence position numbering.
    How each symbol of the alignment is drawn is determined by the
    given :class:`SymbolPlotter` object.

    Parameters
    ----------
    axes : Axes
        A *Matplotlib* axes, that is used as plotting area.
    alignment : Alignment
        The pairwise or multiple sequence alignment to be plotted.
    symbol_plotter : SymbolPlotter
        Defines how the symbols in the alignment are drawn.
    symbols_per_line : int, optional
        The amount of alignment columns that are diplayed per line.
    show_numbers : bool, optional
        If true, the sequence position of the symbols in the last
        alignment column of a line is shown on the right side of the
        plot.
        If the last symbol is a gap, the position of the last actual
        symbol before this gap is taken.
        If the first symbol did not occur up to this point,
        no number is shown for this line.
        By default the first symbol of a sequence has the position 1,
        but this behavior can be changed using the `number_functions`
        parameter.
    number_size : float, optional
        The font size of the position numbers
    number_functions : list of [(None or Callable(int -> int)], optional
        By default the position of the first symbol in a sequence is 1,
        i.e. the sequence position is the sequence index incremented by
        1.
        The behavior can be changed with this parameter:
        If supplied, the length of the list must match the number of
        sequences in the alignment.
        Every entry is a function that maps a sequence index (*int*) to
        a sequence position (*int*) for the respective sequence.
        A ``None`` entry means, that the default numbering is applied
        for the sequence.
    labels : list of str, optional
        The sequence labels.
        Must be the same size and order as the sequences in the
        alignment.
    label_size : float, optional
        Font size of the labels
    show_line_position : bool, optional
        If true the position within a line is plotted below the
        alignment.
    spacing : float, optional
        The spacing between the alignment lines. 1.0 means that the size
        is equal to the size of a symbol box.
    symbol_spacing : int, optional
        А space is placed between each number of elements desired
        by variable.

    See also
    --------
    plot_alignment_similarity_based
    plot_alignment_type_based

    Notes
    -----
    The labels are placed on the Y-axis of the `axes` and the numbers
    are placed on the Y-axis of the `axes` twin.
    The position within a line is placed on the X-axis of the `axes`.
    Further modification of these can be performed using the usual
    *Matplotlib* API.
    """
    from matplotlib.transforms import Bbox

    if number_functions is None:
        number_functions = [lambda x: x + 1] * len(alignment.sequences)
    else:
        if len(number_functions) != len(alignment.sequences):
            raise ValueError(
                f"The amount of renumbering functions is "
                f"{len(number_functions)} but the amount if sequences in the "
                f"alignment is {len(alignment.sequences)}"
            )
        for i, func in enumerate(number_functions):
            if func is None:
                number_functions[i] = (lambda x: x + 1)

    seq_num = alignment.trace.shape[1]
    seq_len = alignment.trace.shape[0]
    line_count = seq_len // symbols_per_line
    # Only extend line count by 1 if there is a remainder
    # (remaining symbols)
    if seq_len % symbols_per_line != 0:
        line_count += 1

    if symbol_spacing:
        spacing_ratio = symbols_per_line / symbol_spacing
        if spacing_ratio % 1 != 0:
            raise ValueError("symbols_per_line not multiple of symbol_spacing")
        # Initializing symbols_to_print to print symbols_per_line
        # symbols on one line + spacing between symbols
        symbols_to_print = int(spacing_ratio) + symbols_per_line - 1
    else:
        symbols_to_print = symbols_per_line

    ### Draw symbols ###
    x = 0
    y = 0
    y_start = 0
    line_pos = 0
    for i in range(seq_len):
        y = y_start
        for j in range(seq_num):
            bbox = Bbox([[x, y], [x+1, y+1]])
            symbol_plotter.plot_symbol(bbox, alignment, i, j)
            y += 1
        line_pos += 1
        if line_pos >= symbols_to_print:
            line_pos = 0
            x = 0
            y_start += seq_num + spacing
        else:
            x += 1
            if (symbol_spacing
               and (i + 1) % symbol_spacing == 0):
                line_pos += 1
                x += 1

    ### Draw labels ###
    ticks = []
    tick_labels = []
    if labels is not None:
        # Labels at center height of each line of symbols -> 0.5
        y = 0.5
        for i in range(line_count):
            for j in range(seq_num):
                ticks.append(y)
                tick_labels.append(labels[j])
                y += 1
            y += spacing
    axes.set_yticks(ticks)
    axes.set_yticklabels(tick_labels)

    ### Draw numbers  ###
    # Create twin to allow different tick labels on right side
    number_axes = axes.twinx()
    ticks = []
    tick_labels = []
    if show_numbers:
        # Numbers at center height of each line of symbols -> 0.5
        y = 0.5
        for i in range(line_count):
            for j in range(seq_num):
                if i == line_count-1:
                    # Last line -> get number of last column in trace
                    trace_pos = len(alignment.trace) - 1
                else:
                    trace_pos = (i+1) * symbols_per_line - 1
                seq_index = _get_last_valid_index(
                    alignment, trace_pos, j
                )
                # if -1 -> terminal gap
                # -> skip number for this sequence in this line
                if seq_index != -1:
                    # Convert sequence index to position
                    # (default index + 1)
                    number = number_functions[j](seq_index)
                    ticks.append(y)
                    tick_labels.append(str(number))
                y += 1
            y += spacing
    number_axes.set_yticks(ticks)
    number_axes.set_yticklabels(tick_labels)

    axes.set_xlim(0, symbols_to_print)
    # Y-axis starts from top
    lim = seq_num*line_count + spacing*(line_count-1)
    axes.set_ylim(lim, 0)
    number_axes.set_ylim(lim, 0)
    axes.set_frame_on(False)
    number_axes.set_frame_on(False)
    # Remove ticks and set label and number size
    axes.yaxis.set_tick_params(
        left=False, right=False, labelsize=label_size
    )
    number_axes.yaxis.set_tick_params(
        left=False, right=False, labelsize=number_size
    )

    if show_line_position:
        axes.xaxis.set_tick_params(
            top=False, bottom=True, labeltop=False, labelbottom=True
        )
    else:
        axes.xaxis.set_tick_params(
            top=False, bottom=False, labeltop=False, labelbottom=False
        )


def plot_alignment_similarity_based(axes, alignment, symbols_per_line=50,
                                    show_numbers=False, number_size=None,
                                    number_functions=None,
                                    labels=None, label_size=None,
                                    show_line_position=False,
                                    spacing=1,
                                    color=None, cmap=None, matrix=None,
                                    color_symbols=False, symbol_spacing=None,
                                    symbol_size=None, symbol_param=None):
    r"""
    Plot a pairwise or multiple sequence alignment highlighting
    the similarity per alignment column.

    This function works like :func:`plot_alignment()` with a
    :class:`SymbolPlotter`, that colors the symbols based on the
    similarity with the other symbols in the same column.
    The color intensity (or colormap value, respectively) of a symbol
    scales with similarity of the respective symbol to the other symbols
    in the same alignment column.

    Parameters
    ----------
    axes : Axes
        A *Matplotlib* axes, that is used as plotting area.
    alignment : Alignment
        The pairwise or multiple sequence alignment to be plotted.
        The alphabet of each sequence in the alignment must be the same.
    symbol_plotter : SymbolPlotter
        Defines how the symbols in the alignment are drawn.
    symbols_per_line : int, optional
        The amount of alignment columns that are diplayed per line.
    show_numbers : bool, optional
        If true, the sequence position of the symbols in the last
        alignment column of a line is shown on the right side of the
        plot.
        If the last symbol is a gap, the position of the last actual
        symbol before this gap is taken.
        If the first symbol did not occur up to this point,
        no number is shown for this line.
        By default the first symbol of a sequence has the position 1,
        but this behavior can be changed using the `number_functions`
        parameter.
    number_size : float, optional
        The font size of the position numbers
    number_functions : list of [(None or Callable(int -> int)], optional
        By default the position of the first symbol in a sequence is 1,
        i.e. the sequence position is the sequence index incremented by
        1.
        The behavior can be changed with this parameter:
        If supplied, the length of the list must match the number of
        sequences in the alignment.
        Every entry is a function that maps a sequence index (*int*) to
        a sequence position (*int*) for the respective sequence.
        A `None` entry means, that the default numbering is applied
        for the sequence.
    labels : list of str, optional
        The sequence labels.
        Must be the same size and order as the sequences in the
        alignment.
    label_size : float, optional
        Font size of the labels
    show_line_position : bool, optional
        If true the position within a line is plotted below the
        alignment.
    spacing : float, optional
        The spacing between the alignment lines. 1.0 means that the size
        is equal to the size of a symbol box.
    color : tuple or str, optional
        A *Matplotlib* compatible color.
        If this parameter is given, the box color in an interpolated
        value between white and the given color,
        or, if `color_symbols` is set to true, between the given color
        and black.
        The interpolation percentage is given by the average normalized
        similarity.
    cmap : Colormap or str, optional
        The boxes (or symbols, if `color_symbols` is set) are
        colored based on the normalized similarity value on the
        given *Matplotlib* Colormap.
    matrix : SubstitutionMatrix
        The substitution matrix used to determine the similarity
        of two symbols. By default an identity matrix is used, i.e.
        only match and mismatch is distinguished.
    color_symbols : bool, optional
        If true, the symbols themselves are colored.
        If false, the symbols are black, and the boxes behind the
        symbols are colored.
    symbol_size : float, optional
        Font size of the sequence symbols.
    symbol_param : dict
        Additional parameters that is given to the
        :class:`matplotlib.Text` instance of each symbol.
    symbol_spacing : int, optional
        А space is placed between each number of elements desired
        by variable.

    See also
    --------
    plot_alignment
    LetterSimilarityPlotter

    Notes
    -----
    For determination of the color, a measure called
    *average normalized similarity* is used.

    The *normalized similarity* of one symbol *a* to another symbol *b*
    (both in aphabet *X*) is defined as

    .. math:: S_{norm}(a,b) = \frac{S(a,b) - \min\limits_x(S(a,x))} {\max\limits_x(S(a,x)) - \min\limits_x(S(a,x))}

    .. math:: a,b,x \in X

    where *S(x,y)* is the similarity score of the two symbols
    *x* and *y* described in the substitution matrix.
    The similarity *S(x,-)* is always 0.
    As the normalization is conducted only with respect to *a*,
    the *normalized similarity* is not commutative.

    The *average normalized similarity* of a symbol *a* is
    determined by averaging the normalized similarity over each
    symbol *b*\ :sub:`i` in the same alignment column.

    .. math:: S_{norm,av}(a) = \frac{1}{n-1} \left[\left(\sum\limits_{i=1}^n S_{norm}(a,b_i)\right) - S_{norm}(a,a)\right]

    The normalized similarity of *a* to itself is subtracted,
    because *a* does also occur in *b*\ :sub:`i`.
    """
    symbol_plotter = LetterSimilarityPlotter(
        axes, matrix=matrix, font_size=symbol_size, font_param=symbol_param,
        color_symbols=color_symbols
    )
    if color is not None or cmap is not None:
        symbol_plotter.set_color(color=color, cmap=cmap)
    plot_alignment(
        axes=axes, alignment=alignment, symbol_plotter=symbol_plotter,
        symbols_per_line=symbols_per_line,
        show_numbers=show_numbers, number_size=number_size,
        number_functions=number_functions,
        labels=labels, label_size=label_size,
        show_line_position=show_line_position,
        spacing=spacing, symbol_spacing=symbol_spacing
    )


def plot_alignment_type_based(axes, alignment, symbols_per_line=50,
                              show_numbers=False, number_size=None,
                              number_functions=None,
                              labels=None, label_size=None,
                              show_line_position=False,
                              spacing=1,
                              color_scheme=None, color_symbols=False,
                              symbol_size=None, symbol_param=None,
                              symbol_spacing=None):
    """
    Plot a pairwise or multiple sequence alignment coloring each symbol
    based on the symbol type.

    This function works like :func:`plot_alignment()` with a
    :class:`SymbolPlotter`, that colors the symbols based on a color
    scheme.
    The color intensity (or colormap value, respectively) of a symbol
    scales with similarity of the respective symbol to the other symbols
    in the same alignment column.

    Parameters
    ----------
    axes : Axes
        A *Matplotlib* axes, that is used as plotting area.
    alignment : Alignment
        The pairwise or multiple sequence alignment to be plotted.
        The alphabet of each sequence in the alignment must be the same.
    symbol_plotter : SymbolPlotter
        Defines how the symbols in the alignment are drawn.
    symbols_per_line : int, optional
        The amount of alignment columns that are diplayed per line.
    show_numbers : bool, optional
        If true, the sequence position of the symbols in the last
        alignment column of a line is shown on the right side of the
        plot.
        If the last symbol is a gap, the position of the last actual
        symbol before this gap is taken.
        If the first symbol did not occur up to this point,
        no number is shown for this line.
        By default the first symbol of a sequence has the position 1,
        but this behavior can be changed using the `number_functions`
        parameter.
    number_size : float, optional
        The font size of the position numbers
    number_functions : list of [(None or Callable(int -> int)], optional
        By default the position of the first symbol in a sequence is 1,
        i.e. the sequence position is the sequence index incremented by
        1.
        The behavior can be changed with this parameter:
        If supplied, the length of the list must match the number of
        sequences in the alignment.
        Every entry is a function that maps a sequence index (*int*) to
        a sequence position (*int*) for the respective sequence.
        A `None` entry means, that the default numbering is applied
        for the sequence.
    labels : list of str, optional
        The sequence labels.
        Must be the same size and order as the sequences in the
        alignment.
    label_size : float, optional
        Font size of the labels
    show_line_position : bool, optional
        If true the position within a line is plotted below the
        alignment.
    spacing : float, optional
        The spacing between the alignment lines. 1.0 means that the size
        is equal to the size of a symbol box.
    color_scheme : str or list of (tuple or str), optional
        Either a valid color scheme name
        (e.g. ``"rainbow"``, ``"clustalx"``, ``blossom``, etc.)
        or a list of *Matplotlib* compatible colors.
        The list length must be at least as long as the
        length of the alphabet used by the sequences.
    color_symbols : bool, optional
        If true, the symbols themselves are colored.
        If false, the symbols are black, and the boxes behind the
        symbols are colored.
    symbol_size : float, optional
        Font size of the sequence symbols.
    symbol_param : dict
        Additional parameters that is given to the
        :class:`matplotlib.Text` instance of each symbol.
    symbol_spacing : int, optional
        А space is placed between each number of elements desired
        by variable.

    See also
    --------
    plot_alignment
    LetterTypePlotter
    """
    alphabet = alignment.sequences[0].get_alphabet()
    symbol_plotter = LetterTypePlotter(
        axes, alphabet, font_size=symbol_size, font_param=symbol_param,
        color_symbols=color_symbols, color_scheme=color_scheme
    )
    plot_alignment(
        axes=axes, alignment=alignment, symbol_plotter=symbol_plotter,
        symbols_per_line=symbols_per_line,
        show_numbers=show_numbers, number_size=number_size,
        number_functions=number_functions,
        labels=labels, label_size=label_size,
        show_line_position=show_line_position,
        spacing=spacing, symbol_spacing=symbol_spacing
    )


def _get_last_valid_index(alignment, column_i, seq_i):
    """
    Find the last trace value that belongs to a valid sequence index
    (no gap -> no -1) up to the specified column.
    """
    index_found = False
    while not index_found:
        if column_i == -1:
            # Iterated from column_i back to beyond the beginning
            # and no index has been found
            # -> Terminal gap
            # -> First symbol of sequence has not occured yet
            # -> return -1
            index = -1
            index_found = True
        else:
            index = alignment.trace[column_i, seq_i]
            if index != -1:
                index_found = True
        column_i -= 1
    return index



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__author__ = "Patrick Kunzmann"

import pkgutil
import doctest
from os.path import join
import tempfile
from importlib import import_module
import numpy as np
import pytest
import biotite.structure.io as strucio
from .util import is_not_installed, cannot_import, cannot_connect_to


NCBI_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/"
RCSB_URL = "https://www.rcsb.org/"
UNIPROT_URL = "https://www.uniprot.org/"


@pytest.mark.parametrize("package_name, context_package_names", [
    pytest.param(
        "biotite",
        []
    ),
    pytest.param(
        "biotite.sequence",
        []
    ),
    pytest.param(
        "biotite.sequence.align",
        ["biotite.sequence"]
    ),
    pytest.param(
        "biotite.sequence.phylo",
        ["biotite.sequence"]
    ),
    pytest.param(
        "biotite.sequence.graphics",
        ["biotite.sequence"],
        marks = pytest.mark.skipif(
            cannot_import("matplotlib"), reason="Matplotlib is not installed"
        )
    ),
    pytest.param(
        "biotite.sequence.io",
        ["biotite.sequence"]
    ),
    pytest.param(
        "biotite.sequence.io.fasta",
        ["biotite.sequence"]
    ),
    pytest.param(
        "biotite.sequence.io.fastq",
        ["biotite.sequence"]
    ),
    pytest.param(
        "biotite.sequence.io.genbank",
        ["biotite.sequence", "biotite.database.entrez"],
        marks = pytest.mark.skipif(
            cannot_connect_to(NCBI_URL), reason="NCBI Entrez is not available"
        )
    ),
    pytest.param(
        "biotite.sequence.io.gff",
        ["biotite.sequence", "biotite.sequence.io.fasta"],
        marks = pytest.mark.filterwarnings("ignore:")
    ),
    pytest.param(
        "biotite.structure",
        ["biotite.structure.io", "biotite.structure.info"]
    ),
    pytest.param(
        "biotite.structure.graphics",
        ["biotite.structure"],    
        marks = pytest.mark.skipif(
            cannot_import("matplotlib"), reason="Matplotlib is not installed"
        ),
    ),
    pytest.param(
        "biotite.structure.io",
        ["biotite.structure"]
    ),
    pytest.param(
        "biotite.structure.io.pdb",
        ["biotite.structure", "biotite"]
    ),
    pytest.param(
        "biotite.structure.io.pdbx",
        ["biotite.structure"]
    ),
    pytest.param(
        "biotite.structure.io.pdbqt",
        ["biotite.structure", "biotite.structure.info"]
    ),
    pytest.param(
        "biotite.structure.io.npz",
        ["biotite.structure"]
    ),
    pytest.param(
        "biotite.structure.io.mmtf",
        ["biotite.structure"]
    ),
    pytest.param(
        "biotite.structure.io.mol",
        ["biotite.structure"]
    ),
    pytest.param(
        "biotite.structure.info",
        ["biotite.structure"]
    ),
    pytest.param(
        "biotite.database.entrez",
        [],                           
        marks = pytest.mark.skipif(
            cannot_connect_to(NCBI_URL), reason="NCBI Entrez is not available"
        )
    ),
    pytest.param(
        "biotite.database.rcsb",
        [],
        marks = pytest.mark.skipif(
            cannot_connect_to(RCSB_URL), reason="RCSB PDB is not available"
        )
    ),
    pytest.param(
        "biotite.database.uniprot",
        [],
        marks = pytest.mark.skipif(
            cannot_connect_to(UNIPROT_URL), reason="UniProt is not available"
        )
    ),
    pytest.param(
        "biotite.application",
        ["biotite.application.clustalo", "biotite.sequence"],            
        marks = pytest.mark.skipif(
            is_not_installed("clustalo"), reason="Software is not installed"
        )
    ),
    pytest.param(
        "biotite.application.blast",
        [],
    ),
    # Do not test Muscle due to version clash
    #pytest.param(
    #    "biotite.application.muscle",
    #    ["biotite.sequence"],
    #    marks = pytest.mark.skipif(
    #        is_not_installed("muscle"), reason="Software is not installed")
    #    ),
    pytest.param(
        "biotite.application.clustalo",
        ["biotite.sequence"],
        marks = pytest.mark.skipif(
            is_not_installed("clustalo"), reason="Software is not installed"
        )
    ),
    pytest.param(
        "biotite.application.mafft",
        ["biotite.sequence"],
        marks = pytest.mark.skipif(
            is_not_installed("mafft"), reason="Software is not installed")
        ),
    pytest.param(
        "biotite.application.sra", ["biotite.sequence"],
        marks = pytest.mark.skipif(
            is_not_installed("fasterq-dump"),
            reason="Software is not installed"
        )
    ),
    pytest.param(
        "biotite.application.tantan",
        ["biotite.sequence"],
        marks = pytest.mark.skipif(
            is_not_installed("tantan"), reason="Software is not installed")
        ),
    pytest.param(
        "biotite.application.viennarna",
        ["biotite.sequence"],
        marks = pytest.mark.skipif(
            is_not_installed("RNAfold") | is_not_installed("RNAplot"),
            reason="Software is not installed"
        )
    ),                                      
    pytest.param(
        "biotite.application.dssp",
        ["biotite.structure"],
        marks = pytest.mark.skipif(
            is_not_installed("mkdssp"), reason="Software is not installed"
        )
    ),
    pytest.param(
        "biotite.application.autodock",
        ["biotite.structure", "biotite.structure.info"],
        marks = pytest.mark.skipif(
            is_not_installed("vina"), reason="Software is not installed"
        )
    ),
])
def test_doctest(package_name, context_package_names):
    """
    Run all doctest strings in all Biotite subpackages.
    """
    # Collect all attributes of this package and its subpackages
    # as globals for the doctests
    globs = {}
    #The package itself is also used as context
    for name in context_package_names + [package_name]:
        context_package = import_module(name)
        globs.update(
            {attr : getattr(context_package, attr)
             for attr in dir(context_package)}
        )
    
    # Add fixed names for certain paths
    globs["path_to_directory"]  = tempfile.gettempdir()
    globs["path_to_structures"] = join(".", "tests", "structure", "data")
    globs["path_to_sequences"]  = join(".", "tests", "sequence", "data")
    # Add frequently used modules
    globs["np"] = np
    # Add frequently used objects
    globs["atom_array_stack"] = strucio.load_structure(
        join(".", "tests", "structure", "data", "1l2y.mmtf"),
        include_bonds=True
    )
    globs["atom_array"] = globs["atom_array_stack"][0]
    
    # Adjust NumPy print formatting
    np.set_printoptions(precision=3, floatmode="maxprec_equal")

    # Run doctests
    # This test does not use 'testfile()' or 'testmod()'
    # due to problems with doctest identification for Cython modules
    # More information below
    package = import_module(package_name)
    runner = doctest.DocTestRunner(
        verbose = False,
        optionflags = 
            doctest.ELLIPSIS |
            doctest.REPORT_ONLY_FIRST_FAILURE |
            doctest.NORMALIZE_WHITESPACE
    )
    for test in doctest.DocTestFinder(exclude_empty=False).find(
        package, package.__name__,
        # It is necessary to set 'module' to 'False', as otherwise
        # Cython functions and classes would be falsely identified
        # as members of an external module by 'DocTestFinder._find()'
        # and consequently would be ignored
        #
        # Setting 'module=False' omits this check
        # This check is not necessary as the biotite subpackages
        # ('__init__.py' modules) should only contain attributes, that
        # are part of the package itself.
        module=False,
        extraglobs=globs
    ):
        runner.run(test)
    results = doctest.TestResults(runner.failures, runner.tries)
    try:
        assert results.failed == 0
    except AssertionError:
        print(f"Failing doctest in module {package}")
        raise


# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
This module provides functions for structure superimposition.
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann, Claude J. Rogers"
__all__ = ["superimpose", "superimpose_apply"]

import numpy as np
from .geometry import centroid


def superimpose(fixed, mobile, atom_mask=None):
    """
    Superimpose structures onto a fixed structure.
    
    The superimposition is performed using the Kabsch algorithm
    :footcite:`Kabsch1976, Kabsch1978`, so that the RMSD between the
    superimposed and the fixed structure is minimized.
    
    Parameters
    ----------
    fixed : AtomArray
        The fixed structure.
    mobile: AtomArray or AtomArrayStack
        The structure(s) which is/are superimposed on the `fixed`
        structure. Both, `fixed` and `mobile` should have equal
        annotation arrays and must have equal sizes.
    atom_mask: ndarray, dtype=bool, optional
        If given, only the atoms covered by this boolean mask will be
        considered for superimposition.
        This means that the algorithm will minimize the RMSD based
        on the covered atoms instead of all atoms.
        The returned superimposed structure will contain all atoms
        of the input structure, regardless of this parameter.
    
    Returns
    -------
    fitted : AtomArray or AtomArrayStack
        A copy of the `mobile` structure(s),
        superimposed on the fixed structure.
    transformation : tuple or tuple list
        The tuple contains the transformations that were applied on
        `mobile`. This can be used in `apply_superimposition()` in order
        to transform another AtomArray in the same way.
        The first element contains the translation vector for moving the
        centroid into the origin.
        The second element contains the rotation matrix.
        The third element contains the translation vector for moving the
        structure onto the fixed.
        The three transformations are performed sequentially.
    
    See Also
    --------
    superimpose_apply
    
    Notes
    -----
    The `transformation` tuple can be used in
    :func:`superimpose_apply()` in order to transform another
    :class:`AtomArray` in the same way.
    This can come in handy, in case you want to superimpose two
    structures with different amount of atoms.
    Often the two structures need to be filtered in order to obtain the
    same size and annotation arrays.
    After superimposition the transformation can be applied on the
    original structure using :func:`superimpose_apply()`.
    
    References
    ----------
    
    .. footbibliography::
    
    Examples
    --------
    
    At first two models of a structure are taken and one of them is
    randomly rotated/translated.
    Consequently the RMSD is quite large:
    
    >>> array1 = atom_array_stack[0]
    >>> array2 = atom_array_stack[1]
    >>> array2 = translate(array2, [1,2,3])
    >>> array2 = rotate(array2, [1,2,3])
    >>> print("{:.3f}".format(rmsd(array1, array2)))
    11.260
    
    RMSD decreases after superimposition of only CA atoms:
    
    >>> array2_fit, transformation = superimpose(
    ...     array1, array2, atom_mask=(array2.atom_name == "CA")
    ... )
    >>> print("{:.3f}".format(rmsd(array1, array2_fit)))
    1.961

    RMSD is even lower when all atoms are considered in the
    superimposition:

    >>> array2_fit, transformation = superimpose(array1, array2)
    >>> print("{:.3f}".format(rmsd(array1, array2_fit)))
    1.928
    """

    try:
        m_coord = mobile.coord
        f_coord = fixed.coord
    except AttributeError:  # e.g., user passed ndarray
        raise ValueError("Expected AtomArray or AtomArrayStack inputs")
    mdim = m_coord.ndim
    if f_coord.ndim != 2:
        raise ValueError("Expected fixed array to be an AtomArray")
    if mdim < 2:
        raise ValueError(
            "Expected mobile array to be an AtomArray or AtomArrayStack"
        )
    if mdim == 2:
        # normalize inputs. Fixed coords has shape (n, 3)
        # and mobile has shape (m, n, 3)
        m_coord = m_coord[np.newaxis, ...]

    nmodels = m_coord.shape[0]
    if f_coord.shape[0] != m_coord.shape[1]:
        raise ValueError(
            f"Expected fixed array and mobile array to have the same number "
            f"of atoms, but {f_coord.shape[0]} != {m_coord.shape[1]}"
        )

    if atom_mask is not None:
        # Implicitly this creates array copies
        mob_filtered = m_coord[..., atom_mask, :]
        fix_filtered = f_coord[atom_mask, :]
    else:
        mob_filtered = np.copy(m_coord)
        fix_filtered = np.copy(f_coord)
    
    # Center coordinates at (0,0,0)
    mob_centroid = centroid(mob_filtered)
    fix_centroid = centroid(fix_filtered)
    mob_filtered -= mob_centroid[..., np.newaxis, :]
    fix_filtered -= fix_centroid
    
    superimposed = mobile.copy()
    s_coord = m_coord.copy() - mob_centroid[..., np.newaxis, :]
    # Perform Kabsch algorithm for every model
    transformations = [None] * nmodels
    for i in range(nmodels):
        rotation = _superimpose(fix_filtered, mob_filtered[i])
        s_coord[i] = np.dot(rotation, s_coord[i].T).T
        transformations[i] = (-mob_centroid[i], rotation, fix_centroid)
    s_coord += fix_centroid
    superimposed.coord = s_coord.reshape(mobile.coord.shape)
    if mdim == 2:
        return superimposed, transformations[0]
    return superimposed, transformations


def _superimpose(fix_centered, mob_centered):
    """
    Perform the Kabsch algorithm using only the coordinates.
    """
    # Calculating rotation matrix
    y = mob_centered
    x = fix_centered
    # Calculate covariance matrix
    cov = np.dot(x.T, y)
    v, s, w = np.linalg.svd(cov)
    # Remove possibility of reflected atom coordinates
    if np.linalg.det(v) * np.linalg.det(w) < 0:
        v[:,-1] *= -1
    rotation = np.dot(v,w)
    return rotation


def superimpose_apply(atoms, transformation):
    """
    Superimpose structures using a given transformation tuple.
    
    The transformation tuple is obtained by prior superimposition.
    
    Parameters
    ----------
    atoms : AtomArray
        The structure to apply the transformation on.
    transformation: tuple, size=3
        The transformation tuple, obtained by :func:`superimpose()`.
    
    Returns
    -------
    fitted : AtomArray or AtomArrayStack
        A copy of the `atoms` structure,
        with transformations applied.
    
    See Also
    --------
    superimpose
    """
    trans1, rot, trans2 = transformation
    transformed = atoms.copy()
    transformed.coord += trans1
    transformed.coord = np.dot(rot, transformed.coord.T).T
    transformed.coord += trans2
    return transformed



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

from tempfile import NamedTemporaryFile
import biotite.structure as struc
import biotite.structure.io as strucio
from biotite.structure.io.general import _guess_element
import numpy as np
import glob
import os
import itertools
from os.path import join, basename, splitext
from ..util import data_dir, cannot_import
import pytest


@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
@pytest.mark.parametrize(
    "path", glob.glob(join(data_dir("structure"), "1l2y.*"))
)
def test_loading(path):
    """
    Just check if :func:`load_structure()` does not raise an exception
    and returns an object of appropriate type.
    """
    suffix = splitext(path)[1]
    if suffix in [".trr", ".xtc", ".tng", ".dcd", ".netcdf"]:
        template = strucio.load_structure(
            join(data_dir("structure"), "1l2y.mmtf")
        )
        array = strucio.load_structure(path, template)
    else:
        array = strucio.load_structure(path)
    if suffix == ".gro":
        # The test GRO file contains only a single model,
        # since it is created by Gromacs
        assert isinstance(array, struc.AtomArray)
    else:
        assert isinstance(array, struc.AtomArrayStack)


@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
def test_loading_template_with_trj():
    """
    Check if :func:`load_structure()` using a trajectory file does not
    raise an exception and returns an object of appropriate type and
    shape.
    """
    template = join(data_dir("structure"), "1l2y.pdb")
    trajectory = join(data_dir("structure"), "1l2y.xtc")
    stack = strucio.load_structure(trajectory, template)
    assert isinstance(stack, struc.AtomArrayStack)
    assert len(stack) > 1


@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
def test_loading_with_extra_args():
    """
    Check if :func:`load_structure()` witt optional arguments does not
    raise an exception and returns an object of appropriate type.
    """
    template = join(data_dir("structure"), "1l2y.pdb")
    trajectory = join(data_dir("structure"), "1l2y.xtc")

    # test if arguments are passed to text files as get_structure arg
    structure = strucio.load_structure(template, extra_fields=["b_factor"])
    assert "b_factor" in structure.get_annotation_categories()

    # test if arguments are passed to read for trajectories
    stack = strucio.load_structure(
        trajectory, template=structure[0], start=5, stop=6
    )
    assert len(stack) == 1

    # loading should fail with wrong arguments
    with pytest.raises(TypeError):
        strucio.load_structure(template, start=2)
    
    # test if atom_i argument is passed to templates
    stack = strucio.load_structure(trajectory, template, atom_i=[1, 2])
    assert stack.shape[1] == 2


@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
@pytest.mark.parametrize(
    "suffix",
    ["pdb", "cif", "gro", "pdbx", "mmtf",
     "trr", "xtc", "tng", "dcd", "netcdf"]
)
def test_saving(suffix):
    """
    Check if loading a structure from a file written via
    :func:`save_structure()` gives the same result as the input to
    :func:`save_structure()`.
    """
    path = join(data_dir("structure"), "1l2y.mmtf")
    ref_array = strucio.load_structure(path)
    if suffix in ("trr", "xtc", "tng", "dcd", "netcdf"):
        # Reading a trajectory file requires a template
        template = path
    else:
        template = None

    temp = NamedTemporaryFile("w", suffix=f".{suffix}", delete=False)
    strucio.save_structure(temp.name, ref_array)
    temp.close()
    
    test_array = strucio.load_structure(temp.name, template)
    os.remove(temp.name)

    for category in ref_array.get_annotation_categories():
        if category == "chain_id" and suffix == "gro":
            # The chain ID is not written to GRO files
            continue
        assert test_array.get_annotation(category).tolist() \
            ==  ref_array.get_annotation(category).tolist()
    assert test_array.coord.flatten().tolist() == pytest.approx(
            ref_array.coord.flatten().tolist(), abs=1e-2
    )


@pytest.mark.skipif(
    cannot_import("mdtraj"),
    reason="MDTraj is not installed"
)
@pytest.mark.parametrize(
    "suffix",
    ["pdb", "cif", "gro", "pdbx", "mmtf",
     "trr", "xtc", "tng", "dcd", "netcdf"]
)
def test_saving_with_extra_args(suffix):
    """
    Test if giving a wrong optional parameter to
    :func:`save_structure()` raises a :class:`TypeError`
    """
    array = strucio.load_structure(join(data_dir("structure"), "1l2y.mmtf"))
    temp = NamedTemporaryFile("w+", suffix=f".{suffix}")
    with pytest.raises(TypeError):
        strucio.save_structure(
            temp.name, array, answer=42
        )
    temp.close()


def test_small_molecule():
    """
    Check if loading a small molecule file written via
    :func:`save_structure()` gives the same result as the input to
    :func:`save_structure()`.
    """
    path = join(data_dir("structure"), "molecules", "TYR.sdf")
    ref_array = strucio.load_structure(path)
    temp = NamedTemporaryFile("w", suffix=".sdf", delete=False)
    strucio.save_structure(temp.name, ref_array)
    temp.close()
    
    test_array = strucio.load_structure(temp.name)
    os.remove(temp.name)

    assert test_array == ref_array


@pytest.mark.parametrize(
    "name,expected",
    [("CA", "C"),
     ("C", "C"),
     ("CB", "C"),
     ("OD1", "O"),
     ("HD21", "H"),
     ("1H", "H"),
     ("CL", "C"),
     ("HE", "H"),
     ("SD", "S"),
     ("NA", "N"),
     ("NX", "N"),
     ("BE", "BE"),
     ("BEA", "BE"),
     ("K", "K"),
     ("KA", "K"),
     ("QWERT", "")]
)
def test_guess_element(name, expected):
    """
    Check if elements are correctly guessed based on known examples.
    Elements are automatically guessed in GRO and PDB files where the
    *element* column is missing.
    """
    result = _guess_element(name)
    assert result == expected



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

"""
A subpackage for handling molecular structures.

In this context an atom is described by two kinds of attributes: the
coordinates and the annotations. The annotations include information
about polypetide chain id, residue id, residue name, hetero atom
information, atom name and optionally more. The coordinates are a
`NumPy` float :class:`ndarray` of length 3, containing the x, y and z
coordinates.

An :class:`Atom` contains data for a single atom, it stores the
annotations as scalar values and the coordinates as length 3
:class:`ndarray`.

An :class:`AtomArray` stores data for an entire structure model
containing *n* atoms.
Therefore the annotations are represented as :class:`ndarray` objects of
length *n*, the so called annotation arrays.
The coordinates are a *(n x 3)* :class:`ndarray`.

An :class:`AtomArrayStack` stores data for *m* models, where each model
contains the same atoms at different positions.
Hence, the annotation arrays are represented as :class:`ndarray` objects
of length *n* like the :class:`AtomArray`, while the coordinates are a
*(m x n x 3)* :class:`ndarray`.

Like an :class:`AtomArray` can be iterated to obtain :class:`Atom`
objects, an :class:`AtomArrayStack` yields :class:`AtomArray` objects.
All three types must not be subclassed.

The following annotation categories are mandatory:

=========  ===========  =================  =======================================
Category   Type         Examples           Description
=========  ===========  =================  =======================================
chain_id   string (U4)  'A','S','AB', ...  Polypeptide chain
res_id     int          1,2,3, ...         Sequence position of residue
ins_code   string (U1)  '', 'A','B',..     PDB insertion code (iCode)
res_name   string (U3)  'GLY','ALA', ...   Residue name
hetero     bool         True, False        False for ``ATOM``, true for ``HETATM``
atom_name  string (U6)  'CA','N', ...      Atom name
element    string (U2)  'C','O','SE', ...  Chemical Element
=========  ===========  =================  =======================================

For all :class:`Atom`, :class:`AtomArray` and :class:`AtomArrayStack`
objects these annotations are initially set with default values.
Additionally to these annotations, an arbitrary amount of annotation
categories can be added via :func:`add_annotation()` or
:func:`set_annotation()`.
The annotation arrays can be accessed either via the method
:func:`get_annotation()` or directly (e.g. ``array.res_id``).

The following annotation categories are optionally used by some
functions:

=========  ===========  =================   ============================
Category   Type         Examples            Description
=========  ===========  =================   ============================
atom_id    int          1,2,3, ...          Atom serial number
b_factor   float        0.9, 12.3, ...      Temperature factor
occupancy  float        .1, .3, .9, ...     Occupancy
charge     int          -2,-1,0,1,2, ...    Electric charge of the atom
=========  ===========  =================   ============================

For each type, the attributes can be accessed directly.
Both :class:`AtomArray` and :class:`AtomArrayStack` support
*NumPy* style indexing.
The index is propagated to each attribute.
If a single integer is used as index,
an object with one dimension less is returned
(:class:`AtomArrayStack` -> :class:`AtomArray`,
:class:`AtomArray` -> :class:`Atom`).
If a slice, index array or a boolean mask is given, a substructure is
returned
(:class:`AtomArrayStack` -> :class:`AtomArrayStack`,
:class:`AtomArray` -> :class:`AtomArray`)
As in *NumPy*, these are not necessarily deep copies of the originals:
The attributes of the sliced object may still point to the original
:class:`ndarray`.
Use the :func:`copy()` method if a deep copy is required.

Bond information can be associated to an :class:`AtomArray` or
:class:`AtomArrayStack` by setting the ``bonds`` attribute with a
:class:`BondList`.
A :class:`BondList` specifies the indices of atoms that form chemical
bonds.
Some functionalities require that the input structure has an associated
:class:`BondList`.
If no :class:`BondList` is associated, the ``bonds`` attribute is
``None``.

Based on the implementation in *NumPy* arrays, this package furthermore
contains a comprehensive set of functions for structure analysis,
manipulation and visualization.

The universal length unit in this package is Å.
"""

__name__ = "biotite.structure"
__author__ = "Patrick Kunzmann"

from .atoms import *
from .bonds import *
from .box import *
from .celllist import *
from .charges import *
from .compare import *
from .density import *
from .dotbracket import *
from .error import *
from .filter import *
from .geometry import *
from .hbond import *
from .integrity import *
from .mechanics import *
from .molecules import *
from .pseudoknots import *
from .rdf import *
from .residues import *
from .chains import *
from .sasa import *
from .sse import *
from .superimpose import *
from .transform import *
from .basepairs import *
# util is used internally



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

import itertools
import pytest
import numpy as np
import biotite.sequence as seq


test_cases = {
    "A" : [0],
    "D" : [3],
    "ABC" : [0,1,2,],
    "ABAFF" : [0,1,0,5,5]
}


@pytest.fixture
def alphabet_symbols():
    return "ABCDEF"


@pytest.mark.parametrize(
    "symbols, exp_code, use_letter_alphabet",
    zip(
        list(test_cases.keys()  ) * 2,
        list(test_cases.values()) * 2,
        [False] * len(test_cases) + [True] * len(test_cases)
    )
)
def test_encoding(alphabet_symbols, symbols, exp_code, use_letter_alphabet):
    if use_letter_alphabet:
        alph = seq.LetterAlphabet(alphabet_symbols)
    else:
        alph = seq.Alphabet(alphabet_symbols)
    
    if len(symbols) == 1:
        assert alph.encode(symbols[0]) == exp_code[0]
    else:
        assert list(alph.encode_multiple(symbols)) == list(exp_code)


@pytest.mark.parametrize(
    "exp_symbols, code, use_letter_alphabet",
    zip(
        list(test_cases.keys()  ) * 2,
        list(test_cases.values()) * 2,
        [False] * len(test_cases) + [True] * len(test_cases)
    )
)
def test_decoding(alphabet_symbols, exp_symbols, code, use_letter_alphabet):
    if use_letter_alphabet:
        alph = seq.LetterAlphabet(alphabet_symbols)
    else:
        alph = seq.Alphabet(alphabet_symbols)
    
    code = np.array(code, dtype=np.uint8)
    if len(code) == 1:
        assert alph.decode(code[0]) == exp_symbols[0]
    else:
        assert list(alph.decode_multiple(code)) == list(exp_symbols)


@pytest.mark.parametrize(
    "use_letter_alphabet, is_single_val",
    itertools.product(
        [False, True], [False, True]
    )
)
def test_error(alphabet_symbols, use_letter_alphabet, is_single_val):
    if use_letter_alphabet:
        alph = seq.LetterAlphabet(alphabet_symbols)
    else:
        alph = seq.Alphabet(alphabet_symbols)

    if is_single_val:
        with pytest.raises(seq.AlphabetError):
            alph.encode("G")
        with pytest.raises(seq.AlphabetError):
            alph.encode(42)
        with pytest.raises(seq.AlphabetError):
            alph.decode(len(alphabet_symbols))
        with pytest.raises(seq.AlphabetError):
            alph.decode(-1)
    else:
        with pytest.raises(seq.AlphabetError):
            alph.encode_multiple("G")
        with pytest.raises(seq.AlphabetError):
            alph.encode_multiple([42])
        with pytest.raises(seq.AlphabetError):
            alph.decode_multiple(np.array([len(alphabet_symbols)]))
        with pytest.raises(seq.AlphabetError):
            alph.decode_multiple(np.array([-1]))


@pytest.mark.parametrize(
    "symbols",
    ["ABC", b"ABC", ["A","B","C"],
     np.array(["A","B","C"]), np.array([b"A",b"B",b"C"])]
)
def test_input_types(alphabet_symbols, symbols):
    """
    'LetterAlphabet' handles different input iterable types in different
    ways.
    Assert that all ways work.
    """
    alph = seq.LetterAlphabet(alphabet_symbols)
    code = alph.encode_multiple(symbols)
    conv_symbols = alph.decode_multiple(code)
    
    
    if isinstance(symbols, bytes):
        symbols = symbols.decode("ASCII")
    assert list(conv_symbols) == list(
        [symbol.decode("ASCII") if isinstance(symbol, bytes) else symbol
         for symbol in symbols]
    )


@pytest.mark.parametrize("use_letter_alphabet", [False, True])
def test_length(alphabet_symbols, use_letter_alphabet):
    if use_letter_alphabet:
        alph = seq.LetterAlphabet(alphabet_symbols)
    else:
        alph = seq.Alphabet(alphabet_symbols)
    assert len(alph) == len(alphabet_symbols)


@pytest.mark.parametrize("use_letter_alphabet", [False, True])
def test_contains(alphabet_symbols, use_letter_alphabet):
    if use_letter_alphabet:
        alph = seq.LetterAlphabet(alphabet_symbols)
    else:
        alph = seq.Alphabet(alphabet_symbols)
    assert "D" in alph

@pytest.mark.parametrize(
    "source_alph_symbols, target_alph_symbols", 
    [
        ("A", "AB"),
        (["foo", "bar"], ["bar", "foo", 42]),
        ("ACGT", "AGTC"),
        ("ACGT", "ACGNT"),
        (np.arange(0, 1000), np.arange(999, -1, -1)),
    ]
)
def test_alphabet_mapper(source_alph_symbols, target_alph_symbols):
    CODE_LENGTH = 10000
    source_alph = seq.Alphabet(source_alph_symbols)
    target_alph = seq.Alphabet(target_alph_symbols)
    mapper = seq.AlphabetMapper(source_alph, target_alph)
    
    ref_sequence = seq.GeneralSequence(source_alph)
    np.random.seed(0)
    ref_sequence.code = np.random.randint(
        len(source_alph), size=CODE_LENGTH, dtype=int
    )

    test_sequence = seq.GeneralSequence(target_alph)
    test_sequence.code = mapper[ref_sequence.code]

    assert test_sequence.symbols == ref_sequence.symbols



# This source code is part of the Biotite package and is distributed
# under the 3-Clause BSD License. Please see 'LICENSE.rst' for further
# information.

__name__ = "biotite.database"
__author__ = "Patrick Kunzmann"
__all__ = ["RequestError"]


class RequestError(Exception):
    """
    Indicates that the database returned a response with an error
    message or other malformed content.
    """
    pass